# This program is public domain
import os
import numpy as np
from copy import copy

from reductus.dataflow.automod import cache, nocache, module

# TODO: maybe bring back formula to show the math of each step
# TODO: what about polarized data?

@module
def nop(data):
    """
    No operation.

    **Inputs**

    data (refldata[]): Input data

    **Returns**

    output (refldata[]): Unaltered data

    2015-12-31 Paul Kienzle
    """
    return data

@module
def ncnr_load(filelist=None, check_timestamps=True):
    """
    Load a list of nexus files from the NCNR data server.

    **Inputs**

    filelist (fileinfo[]): List of files to open.

    check_timestamps (bool): verify that timestamps on file match request

    **Returns**

    output (refldata[]): All entries of all files in the list.

    2016-06-29 Brian Maranville
    | 2017-08-21 Brian Maranville Change to refldata, force cache invalidate
    | 2018-06-18 Brian Maranville Change to nexusref to ignore areaDetector
    | 2018-12-10 Brian Maranville get_plottable routines moved to python data container from js
    | 2020-03-03 Paul Kienzle Just load.  Don't even compute divergence
    """
    # NB: used mainly to set metadata for processing, so keep it minimal
    # TODO: make a metadata loader that does not send all data to browser
    # NB: Fileinfo is a structure with
    #     { path: "location/on/server", mtime: timestamp }
    from .load import url_load_list

    datasets = []
    for data in url_load_list(filelist, check_timestamps=check_timestamps):
        datasets.append(data)
    return datasets

@module
def dark_current(data, poly_coeff=[0], poly_cov=None):
    r"""
    Correct for the dark current, which is the average number of
    spurious counts per minute of measurement on each detector channel.

    **Inputs**

    data (refldata[]) : data to scale

    poly_coeff {Polynomial coefficients of dark current vs slit1} (float[])
    : Polynomial coefficients (highest order to lowest) representing the dark current as a function of slit 1 opening. Units in counts/(minute . mm ^ N), for the coefficient of (slit1 ^ N). Default is [0].

    poly_cov {Polynomial covariance matrix} (float[])
    : Flattened covariance matrix for polynomial coefficients if error propagation is desired. For an order N polynomial, must have size N^2. If left blank, no error propagation will occur.

    **Returns**

    output (refldata[]): Dark current subtracted data.

    darkcurrent (refldata[]): Dark current that was subtracted (for plotting)

    | 2020-03-04 Paul Kienzle
    | 2020-03-12 Paul Kienzle Add slit 1 dependence for DC rate
    | 2021-06-11 David Hoogerheide generalize to refldata, prevent either adding counts or oversubtracting
    | 2021-06-13 David Hoogerheide add dark current output
    | 2021-06-14 David Hoogerheide change to polynomial input and add error propagation
    """

    # TODO: datatype hierarchy: accepts any kind of refldata

    from reductus.dataflow.lib.uncertainty import Uncertainty as U

    datasets = list()
    dcs = list()

    order = len(poly_coeff)

    for d in data:
        dcdata = copy(d)                    # hackish way to get dark current counts
        dcdata.detector = copy(d.detector)

        # calculate rate and Jacobian at each point
        rate = np.polyval(poly_coeff, dcdata.slit1.x)

        # error propagation
        rate_var = np.zeros_like(rate)
        if poly_cov is not None:
            poly_cov = np.array(poly_cov).reshape((order, order))
            for i, s1 in enumerate(dcdata.slit1.x):
                J = np.array([s1**float(i) for i in range(0, order)[::-1]])
                rate_var[i] = np.dot(J.T, np.dot(poly_cov, J))
        dc = dcdata.monitor.count_time*(rate/60.)
        dc[dc < 0] = 0.0                            # do not allow addition of dark counts from negative rates
        dc_var = rate_var * (dcdata.monitor.count_time/60.)**2

        # condition dark counts to the correct dimensionality
        ndetectordims = np.ndim(d.detector.counts)
        dc = np.expand_dims(dc, tuple(range(1, ndetectordims)))
        dc_var = np.expand_dims(dc_var, tuple(range(1, ndetectordims)))
        dcdata.detector.counts = np.ones_like(dcdata.detector.counts) * dc  # should preserve dimensionality correctly
        dcdata.detector.counts_variance = np.ones_like(dcdata.detector.counts_variance) * dc_var  # should preserve dimensionality correctly

        detcounts = U(d.detector.counts, d.detector.counts_variance)
        darkcounts = U(dcdata.detector.counts, dcdata.detector.counts_variance)

        detdarkdiff = detcounts - darkcounts

        d.detector.counts, d.detector.counts_variance = detdarkdiff.x, detdarkdiff.variance
        d.detector.counts[d.detector.counts < 0] = 0.0
        
        # only renormalize if apply_norm has already populated d.normbase, i.e. if it's a standalone module
        if d.normbase is not None:
            d = normalize(d, d.normbase)
            dcdata = normalize(dcdata, dcdata.normbase)

        # create outputs
        datasets.append(d)
        dcs.append(dcdata)

    return datasets, dcs

@module
def fit_dead_time(data, source='detector', mode='auto'):
    """
    Fit detector dead time constants (paralyzing and non-paralyzing) from
    measurement of attenuated and unattenuated data for a range of count rates.

    **Inputs**

    data (refldata[]): Data sets with different attenuation levels

    source (opt:detector|monitor): Tube that is being measured

    mode (opt:P|NP|mixed|auto): Dead-time mode

    **Returns**

    dead_time (deadtime): Dead time constants, attenuator estimate and beam rate

    2015-12-17 Paul Kienzle
    """
    from .deadtime import fit_dead_time

    dead_time = fit_dead_time(data, source=source, mode=mode)
    return dead_time



@module
def monitor_dead_time(data, dead_time, nonparalyzing=0.0, paralyzing=0.0):
    """
    Correct the monitor dead time from the fitted dead time.

    The deadtime constants are chosen as follows:

    #. If either *tau_NP* and *tau_P* are non-zero, then use them.
    #. If the dead time terminal is attached to a dead time fit, use it.
    #. If the dead time constants are given in the data file, then use them.
    #. Otherwise don't do any dead time correction.

    **Inputs**

    data (refldata) : Uncorrected data

    dead_time (deadtime?) : Output of dead time estimator

    nonparalyzing (float:us<0,inf>) : non-paralyzing dead time constant

    paralyzing (float:us<0,inf>) : paralyzing dead time constant

    **Returns**

    output (refldata): Dead-time corrected data

    2015-12-17 Paul Kienzle
    """
    from .deadtime import apply_monitor_dead_time

    data = copy(data)
    data.monitor = copy(data.monitor)
    if nonparalyzing != 0.0 or paralyzing != 0.0:
        apply_monitor_dead_time(data, tau_NP=nonparalyzing,
                                tau_P=paralyzing)
    elif dead_time is not None:
        apply_monitor_dead_time(data, tau_NP=dead_time.tau_NP,
                                tau_P=dead_time.tau_P)
    elif data.monitor.deadtime is not None and np.isfinite(data.monitor.deadtime).all():
        try:
            tau_NP, tau_P = data.monitor.deadtime
        except Exception:
            tau_NP, tau_P = data.monitor.deadtime, 0.0
        apply_monitor_dead_time(data, tau_NP=tau_NP, tau_P=tau_P)
    else:
        pass  # no deadtime correction parameters available.

    return data


@module
def detector_dead_time(data, dead_time, nonparalyzing=0.0, paralyzing=0.0):
    """
    Correct the detector dead time from the fitted dead time.

    If *tau_NP* and *tau_P* are non-zero, then use them.  If a dead_time
    fit result is supplied, then use it.  If the dead time constants are
    given in the data file, then use them.  Otherwise don't do any
    dead time correction.

    **Inputs**

    data (refldata) : Uncorrected data

    dead_time (deadtime?) : Output from dead time estimator

    nonparalyzing (float:us<0,inf>) : non-paralyzing dead time constant

    paralyzing (float:us<0,inf>) : paralyzing dead time constant

    **Returns**

    output (refldata): Dead-time corrected data

    2016-03-21 Paul Kienzle
    """
    from .deadtime import apply_detector_dead_time

    data = copy(data)
    if nonparalyzing != 0.0 or paralyzing != 0.0:
        data.detector = copy(data.detector)
        apply_detector_dead_time(data, tau_NP=nonparalyzing,
                                 tau_P=paralyzing)
    elif dead_time is not None:
        data.detector = copy(data.detector)
        apply_detector_dead_time(data, tau_NP=dead_time.tau_NP,
                                 tau_P=dead_time.tau_P)
    elif data.detector.deadtime is not None and not np.all(np.isnan(data.detector.deadtime)):
        try:
            tau_NP, tau_P = data.detector.deadtime
        except Exception:
            tau_NP, tau_P = data.detector.deadtime, 0.0
        data.detector = copy(data.detector)
        apply_detector_dead_time(data, tau_NP=tau_NP, tau_P=tau_P)
    else:
        raise ValueError("no valid deadtime provided in file or parameter")

    return data


@module
def monitor_saturation(data):
    """
    Correct the monitor dead time from stored saturation curve.

    **Inputs**

    data (refldata): Uncorrected data

    **Returns**

    output (refldata): Dead-time corrected data

    2017-02-22 Paul Kienzle
    """
    from .deadtime import apply_monitor_saturation
    data = copy(data)
    if getattr(data.monitor, 'saturation', None) is not None:
        data.monitor = copy(data.monitor)
        apply_monitor_saturation(data)
    else:
        data.warn("no monitor saturation for %r"%data.name)
    return data


@module
def detector_saturation(data):
    """
    Correct the detector dead time from stored saturation curve.

    **Inputs**

    data (refldata): Uncorrected data

    **Returns**

    output (refldata): Dead-time corrected data

    2015-12-17 Paul Kienzle
    """
    from .deadtime import apply_detector_saturation
    data = copy(data)
    if getattr(data.detector, 'saturation', None) is not None:
        #print("detector "+str(data.detector.__dict__))
        data.detector = copy(data.detector)
        apply_detector_saturation(data)
    else:
        data.warn("no detector saturation for %r"%data.name)
    return data


@module
def theta_offset(data, offset=0.0):
    """
    Correct the theta offset of the data for a misaligned sample, shifting
    sample and detector angle and updating $q_x$ and $q_z$.

    **Inputs**

    data (refldata) : Uncorrected data

    offset (float:degree) : amount of shift to add to sample angle and subtract
    from detector angle

    **Returns**

    output (refldata): Offset corrected data

    2015-12-17 Paul Kienzle
    """
    from .angles import apply_theta_offset
    data = copy(data)
    data.sample = copy(data.sample)
    data.detector = copy(data.detector)
    data.sample.angle_x = copy(data.sample.angle_x)
    data.detector.angle_x = copy(data.detector.angle_x)
    apply_theta_offset(data, offset)
    return data


@module
def back_reflection(data):
    """
    Reverse the sense of the reflection angles, making positive angles
    negative and vice versa.

    **Inputs**

    data (refldata): Uncorrected data

    **Returns**

    output (refldata): Angle corrected data

    2015-12-17 Paul Kienzle
    """
    from .angles import apply_back_reflection
    data = copy(data)
    data.sample = copy(data.sample)
    data.detector = copy(data.detector)
    data.sample.angle_x = copy(data.sample.angle_x)
    data.detector.angle_x = copy(data.detector.angle_x)
    apply_back_reflection(data)
    return data


@module
def absolute_angle(data):
    """
    Assume all reflection is off the top surface, reversing the sense
    of negative angles.

    **Inputs**

    data (refldata): Uncorrected data

    **Returns**

    output (refldata): Angle corrected data

    2015-12-17 Paul Kienzle
    """
    from .angles import apply_absolute_angle
    data = copy(data)
    data.sample = copy(data.sample)
    data.detector = copy(data.detector)
    data.sample.angle_x = copy(data.sample.angle_x)
    data.detector.angle_x = copy(data.detector.angle_x)
    apply_absolute_angle(data)
    return data

@module
def sample_broadening(data, width=0):
    r"""
    Increase (or decrease) nominal divergence due to the effects of sample
    broadening (or focussing) if this is not supported by the reflectivity
    analysis program.

    **Inputs**

    data (refldata): data without resolution estimate

    width (float) : amount of increased divergence in degrees, using
    1-\ $\sigma$ change in width.  This can be estimated from the FWHM of the
    rocking curve relative to the expected value with no broadening, divided
    by 2.35 to convert FWHM to 1-\ $\sigma$.

    **Returns**

    output (refldata): data with resolution estimate

    2020-05-05 Paul Kienzle
    """
    from .angles import apply_sample_broadening
    if width != 0:
        data = copy(data)
        apply_sample_broadening(data, width)
    return data

@module
def divergence_fb(data, sample_width=None):
    r"""
    Estimate divergence from slit openings.  Does nothing if divergence
    is already defined by the instrument.

    **Inputs**

    data (refldata): data without resolution estimate

    sample_width (float?:<0,inf>) : width of the sample in mm if it acts like a slit.
    By default, this uses the value in the file.

    **Returns**

    output (refldata): data with resolution estimate

    2020-05-05 Paul Kienzle
    """
    from .angles import apply_divergence_front_back
    if data.angular_resolution is None:
        data = copy(data)
        apply_divergence_front_back(data, sample_width)
    return data

@module
def divergence(data, sample_width=None, sample_broadening=0):
    r"""
    Estimate divergence from slit openings.  Does nothing if divergence
    is already defined by the instrument.

    **DEPRECATED** use divergence_fb instead

    **Inputs**

    data (refldata): data without resolution estimate

    sample_width (float?:<0,inf>) : width of the sample in mm if it acts like a slit.
    By default, this uses the value in the file.

    sample_broadening (float) : amount of increased divergence in degrees, using
    1-\ $\sigma$ change in width.  This can be estimated from the FWHM of the
    rocking curve relative to the expected value with no broadening, divided
    by 2.35 to convert FWHM to 1-\ $\sigma$.

    **Returns**

    output (refldata): data with resolution estimate

    2016-06-15 Paul Kienzle
    2020-05-05 Paul Kienzle
    """
    from .angles import apply_divergence_simple, apply_sample_broadening
    if data.angular_resolution is None:
        data = copy(data)
        apply_divergence_simple(data, sample_width)
        apply_sample_broadening(data, sample_broadening)
    return data


#@module
def mask_specular(data):
    """
    Identify and mask out specular points.

    This defines the *mask* attribute of *data* as including all points that
    are not specular or not previously masked.  The points are not actually
    removed from the data, since this operation is done by *join*.

    **Inputs**

    data (refldata) : background data which may contain specular point

    **Returns**

    output (refldata) : masked data

    2015-12-17 Paul Kienzle
    """
    from .background import apply_specular_mask
    data = copy(data)
    apply_specular_mask(data)
    return data

def mask_action(data=None, mask_indices=None, **kwargs):
    """
    Remove data at the indicated indices
    """
    if mask_indices:
        data = copy(data)
        data.apply_mask(mask_indices)
    return data

@module
def mask_points(data, mask_indices=None):
    """
    Identify and mask out user-specified points.

    This defines the *mask* attribute of *data* to include all data
    except those indicated in *mask_indices*.  Any previous mask is cleared.
    The masked data are not actually removed from the data, since this
    operation is done by *join*.

    **Inputs**

    data (refldata) : background data which may contain specular point

    mask_indices (index[]*) : 0-origin data point indices to mask. For example,
    *mask_indices=[1,4,6]* masks the 2nd, 5th and 7th point respectively. Each
    dataset should have its own mask.

    **Returns**

    output (refldata) : masked data

    | 2018-04-30 Brian Maranville
    | 2019-07-02 Brian Maranville: change self.points after mask
    """
    data = copy(data)
    output = mask_action(data=data, mask_indices=mask_indices)
    return output

@module
def mark_intent(data, intent='auto'):
    r"""
    Mark the file type based on the contents of the file, or override.

    *intent* can be 'infer', to guess the intent from the measurement geometry,
    'auto' to use the recorded value for the intent if present, otherwise
    infer it from the geometry, or the name of the intent.

    For inferred intent, it is 'specular' if incident angle matches detector
    angle within 0.1*angular divergence, 'background+' if incident angle is
    greater than detector angle, 'background-' if incident angle is less
    than detector angle, 'slit' if incident and detector angles are zero,
    'rock sample' if only the incident angle changes, 'rock detector' if
    only the detector angle changes, or 'rock qx' if only $Q_x$ is changing
    throughout the scan.

    **Inputs**

    data (refldata) : data file which may or may not have intent marked

    intent (opt:auto|infer|specular|background+\|background-\|slit
    \|rock sample|rock detector|rock qx) : intent to register with the
    datafile, or auto/infer to guess

    **Returns**

    output (refldata) : marked data

    2016-03-20 Paul Kienzle
    """
    from .intent import apply_intent
    data = copy(data)
    apply_intent(data, intent)
    return data

@module
def group_by_intent(data):
    """
    Split a bundle into multiple bundles using intent.

    **Inputs**

    data (refldata[]) : data files with intent marked

    **Returns**

    specular (refldata[]) : specular measurements

    backp {Background+} (refldata[]) : positive offset background measurements

    backm {Background-} (refldata[]) : negative offset background measurements

    intensity (refldata[]) : beam intensity measurements

    rock {Rocking curve} (refldata[]) : rocking curve measurements

    other (refldata[]) : everything else

    2016-07-20 Brian Maranville
    """
    map_intent = {
        'specular': 'specular',
        'intensity': 'intensity',
        'background+': 'backp',
        'background-': 'backm',
        'rock sample': 'rock',
        'rock detector': 'rock',
        'rock qx': 'rock',
        }
    groups = {}
    for intent in set(map_intent.values()):
        groups[intent] = []
    groups['other'] = []
    for d in data:
        #print("intent %s %s"%(d.intent, d.path))
        groups[map_intent.get(d.intent, 'other')].append(d)

    return [groups[intent]
            for intent in 'specular backp backm intensity rock other'.split()]

@module
def extract_xs(data, xs="++"):
    r"""
    Get a polarization cross-section from a bundle

    **Inputs**

    data (refldata[]): data files in of all cross sections

    xs {Cross-section} (opt:++\|--\|+-\|-+\|+\|-\|unpolarized): cross-section to extract

    **Returns**

    output (refldata[]): data matching just that cross-section

    | 2016-05-05 Brian Maranville
    | 2020-03-24 Brian Maranville: added half-pol cross-sections
    """
    # Note: no need to copy data since it is not being modified
    if xs == 'unpolarized':
        xs = ''
    output = [d for d in data if d.polarization == xs]
    return output

@module
def filter(data, key="", comparator="eq", value=None):
    r"""
    Get a subset of the datasets bundle based on the test

    **Inputs**

    data (refldata[]): data files in

    key (str): name to test in the dataset

    value (str?): value to compare

    comparator {Compare operator} (opt:eq|ne|lt|le|gt|ge): comparison operator

    **Returns**

    output (refldata[]): data matching the comparison

    2017-02-24 Brian Maranville
    """

    import operator
    compare_lookup = {
        "==": operator.eq,
        "!=": operator.ne,
        "<": operator.lt,
        "<=": operator.le,
        ">=": operator.ge,
        ">": operator.gt,
        "eq": operator.eq,
        "ne": operator.ne,
        "lt": operator.lt,
        "le": operator.le,
        "ge": operator.ge,
        "gt": operator.gt,

    }
    compare_op = compare_lookup[comparator]

    return [d for d in data if hasattr(d, key) and compare_op(getattr(d, key), value)]

@module
def normalize(data, base='auto'):
    """
    Estimate the detector count rate.

    *base* can be monitor, time, roi, power, or none for no normalization.
    For example, if base='monitor' then the count rate will be counts
    per monitor count.  Note that operations that combine datasets require
    the same normalization on the points.

    If *base* is auto then the default will be chosen, which is 'monitor'
    if the monitor exists, otherwise it is 'time'.  If neither exists
    (not sure that can happen) then the data will be unnormalized.

    The detector region of interest (*roi*) and reactor *power* have not been
    tested and should not be used. The detector efficient, the dead time
    corrections and attenuator scaling have not been applied to the roi
    measurement.  Since the measurement is only useful if a portion of the
    detector is exposed to the incident beam, this corrections will be
    especially important.  In the case where the monitor is unreliable and
    reactor power has been fluctuating, you may be able to estimate the
    incident intensity based on the integrated reactor power. This uses
    a simple average of the reactor power measurements multiplied by the
    measurement time.

    **Inputs**

    data (refldata) : data to normalize

    base {Normalize by} (opt:auto|monitor|time|roi|power|none)
    : how to convert from counts to count rates

    **Returns**

    output (refldata) : data with count rate rather than counts

    2015-12-17 Paul Kienzle
    2020-03-10 Paul Kienzle auto almost always equals monitor
    """
    # Note: reflpak supported visualization like "counts per 10000 monitor"
    # so that the displayed data looked roughly like the measured data, except
    # all scaled to a common monitor.  This is not available in reductus.

    # TODO: consistent use of data.detector.counts vs. data.v
    # see in particular the detector/monitor dead time, spectral efficiency,
    # dark current, etc.

    from .scale import apply_norm
    data = copy(data)
    apply_norm(data, base)
    return data


@module
def psd_center(data, center=128):
    """
    Set center pixel for the detector.

    **Inputs**

    data (psddata) : data to scale

    center (float) : beam center pixel (should be the same for all datasets)

    **Returns**

    output (psddata) : scaled data

    2020-02-04 Paul Kienzle
    """
    data = copy(data)
    data.detector = copy(data.detector)
    data.detector.center = (center, 0)
    return data

@module
def psd_integrate(
        data, spec_scale=1, spec_pixel=5.,
        left_scale=1., left_pixel=5., right_scale=1., right_pixel=5.,
        min_pixel=5., max_pixel=251., degree=1., mc_samples=1000,
        slices=None, #(0.01, 0.05, 0.10, 0.15),
        ):
    r"""
    Integrate specular and background from psd.

    Specular and background regions are computed from beam divergence and
    pixel width using the following:

        spec = spec scale * divergence + spec pixels
        left = left scale * divergence + left pixels
        right = right scale * divergence + right pixels

    The beam divergence used in the equations above is estimated from
    the slit openings. The specular signal is the sum over pixels
    in [-spec, +spec]. The background signal is determined by fitting
    a polynomial of degree n to the pixels in [-spec - left, -spec)
    and (spec, spec + right), then integrating that polynomial over
    the specular pixels.

    Specular uncertainty comes from simply integrating over the pixels.
    Background uncertainty comes from the uncertainty in the polynomial
    fitting parameters.  It can be estimated using Monte Carlo sampling,
    or by simple Gaussian propagation of uncertainty if mc samples is 0.
    MC estimates are stochastic, so rerunning with a different random
    number sequence will give a different result.  To make the reduction
    reproducible, the number of samples is used as the seed value for the
    random number generator.  To assess the variation in the background
    estimate, try slightly longer sequence lengths.  We have found 100
    samples gives a background estimate that is approximately stable
    (variation in estimate is well within uncertainty). A default value
    of 1000 was chosen because it is reasonably fast and reasonably stable.
    Test on your own data by comparing mc_samples to mc_samples+1

    The residual after subtracting the background estimate is also
    returned. Use this to verify that the integration ranges are
    chosen appropriately.  There is an additional output to show slices
    for selected frames---this will show the quality of background estimate.

    **Inputs**

    data (refldata) : data to integrate

    spec_scale {Spec scale}(float:<0,inf>) : Specular width as a divergence multiple, or zero for fixed width.

    spec_pixel {Spec offset}(float:pixel) : Fixed broadening in pixels (or narrowing if negative)

    left_scale {Back- scale}(float:<0,inf>) : Left background width as a divergence multiple, or zero for fixed width.

    left_pixel {Back- offset}(float:pixel) : Left background shift in pixels.

    right_scale {Back+ scale}(float:<0,inf>) : Right background width as a divergence multiple, or zero for fixed width.

    right_pixel {Back+ offset}(float:pixel) : Right background shift in pixels.

    min_pixel {Left edge}(float:pixel<1,256>) : Left background cutoff pixel.

    max_pixel {Right edge}(float:pixel<1,256>) : Right background cutoff pixel.

    degree {Polynomial degree}(int:<0,9>) : Background polynomial degree.

    mc_samples {MC samples}(int:<0,inf>) : Number of MC samples for uncertainty analysis, or zero for simple gaussian.

    slices {Slice value}(float) : Display data cross-sections at the given values.

    **Returns**

    specular (refldata) : integrated specular

    background (refldata) : integrated background

    residual (psddata) : background subtracted psd data

    sliceplot (plot) : slices plot

    | 2020-02-03 Paul Kienzle
    """
    from .ng7psd import apply_integration
    from reductus.dataflow.data import Plottable

    mc_seed = mc_samples if mc_samples > 0 else None
    #print("slices", slices)
    spec, back, resid, sliceplot = apply_integration(
        data, spec=(spec_scale, spec_pixel),
        left=(left_scale, left_pixel), right=(right_scale, right_pixel),
        pixel_range=(min_pixel, max_pixel),
        degree=degree, mc_samples=mc_samples, seed=mc_seed,
        slices=[slices] if slices is not None else [],
    )

    return spec, back, resid, Plottable(sliceplot)

@module
def rescale(data, scale=1.0, dscale=0.0):
    """
    Rescale the count rate by some scale and uncertainty.

    **Inputs**

    data (refldata) : data to scale

    scale (scale*) : amount to scale, one for each dataset

    dscale {Scale err} (float*:<0,inf>) : scale uncertainty for gaussian error propagation

    **Returns**

    output (refldata) : scaled data

    2015-12-17 Paul Kienzle
    """
    from .scale import apply_rescale
    data = copy(data)
    apply_rescale(data, scale, dscale)
    return data

#@nocache
@module
def join(data, Q_tolerance=0.5, dQ_tolerance=0.002, order='file',
         group_by="polarization", tolerance=None):
    r"""
    Join operates on a list of datasets, returning a list with one dataset,
    or one dataset per polarization state.  When operating on a single
    dataset, it joins repeated points into single points.

    *Qtol* and *dQtol* are scale factors on $\Delta \theta$ used to
    determine whether two angles are equivalent.  For a given tolerance
    $\epsilon_Q, \epsilon_{\Delta Q}$, a point at incident angle
    $\theta_1$ can be joined with one with incident angle $\theta_2$ when
    $|\theta_1 - \theta_2| < \epsilon_Q \cdot \Delta\theta$ and
    $|\Delta\theta_1 - \Delta\theta_2| < \epsilon_{\Delta Q} \cdot \Delta\theta$.
    Values of $\epsilon_Q=0.5$ and $\epsilon_{\Delta Q}=0.002$ work well in
    practice. If the tolerances are both 0 then join is performed against
    the desired positions rather than the actual positions; this more
    closely corresponds with user intent.

    The join algorithm is greedy, so if you have a sequence of points with
    individual separation less than $\epsilon\cdot\Delta\theta$ but total
    spread greater than $\epsilon\cdot\Delta\theta$, they will be joined
    into multiple points with the final with the final point having worse
    statistics than the prior points.  Join operates on one dimension at
    a time, first grouping points with common $\Delta\theta$, then joining
    points within each $\Delta\theta$ by common $\theta_\text{incident}$,
    then by common $\theta_\text{detector}$.  This algorithm should work
    well enough on the common reflectometry scans, but it may fail for example
    if applied to a set of $Q_x$ scans with different $Q_z$ values.

    *order* is the sort order of the files that are joined.  The first
    file in the sorted list determines the metadata such as the base
    file name for the joined file.

    The joined datasets will be sorted as appropriate for the the
    measurement intent.  Masked points will be removed.

    Data must be normalized before join.

    **Inputs**

    data (refldata[]) : data to join

    Q_tolerance (float:1-sigma<0,inf>) : allowed separation between points
    while still joining them to a single point; this is relative to the angular
    resolution and wavelength dispersion of each point

    dQ_tolerance (float:1-sigma<0,inf>) : allowed difference in resolution
    between combined points; this is relative to the angular resolution and
    wavelength dispersion of each point

    order (opt:file|time|theta|slit|none) : order determines which file is the
    base file, supplying the metadata for the joined set

    group_by (opt:polarization|probe|entry|filenumber|instrument|intent|sample.name|sample.description) : key by which the files are grouped prior to join

    tolerance(float?:1-sigma<0,inf>) : **deprecated** value for Qtol and dQtol;
    ignored if the value is None or not specified.

    **Returns**

    output (refldata[]) : joined data

    | 2017-02-09 Paul Kienzle: split tolerance into Qtol and dQtol
    | 2017-05-05 Paul Kienzle: consistent sort order for outputs
    | 2017-07-03 Brian Maranville: rearrange to group by Ti, Td before dT, dQ
    | 2018-05-14 Brian Maranville: group by Qx first for all rocking curves
    | 2020-10-14 Paul Kienzle fixed uncertainty for time normalized data
    | 2020-12-15 Brian Maranville added roi_counts and source_power to columns
    """
    from .joindata import sort_files, join_datasets
    from .util import group_by_key
    # No copy necessary; join is never in-place.

    # TODO: parse **deprecated** in automod and hide deprecated inputs on ui
    if tolerance is not None:
        Q_tolerance = dQ_tolerance = tolerance
    datasets = [v for k, v in sorted(group_by_key(group_by, data).items())]
    output = []
    for group in datasets:
        group = sort_files(group, order)
        result = join_datasets(group, Q_tolerance, dQ_tolerance)
        output.append(result)
    return output

@module
def mix_cross_sections(data, mix_sf=False, mix_nsf=False):
    """
    Mix (combine) cross-sections, usually to improve statistics when cross-sections
    are expected to be indistinguishable in the model (e.g. spin-flip when no chirality)
    Typically this is done after load and before "join"

    All inputs are passed to the output, and in addition:

    When *mix_sf* is enabled, all input datasets with polarization "-+" will be copied and
    added to the output with polarization = "+-", and vice-versa for "+-" inputs.

    When *mix_nsf* is enabled, all input datasets with polarization "++" will be copied and
    added to the output with polarization = "--", and similarly "--" inputs sent to "++"

    **Inputs**

    data (refldata[]) : datasets in

    mix_sf {Mix Spin-Flip?} (bool) : Perform mixing on spin-flip cross-sections,
    i.e. "+-" and "-+"

    mix_nsf {Mix Non-Spin-Flip?} (bool) : Perform mixing on spin-flip cross-sections,
    i.e. "++" and "--" or "+" and "-"

    **Returns**

    output (refldata[]) : relabeled and copied datasets (around twice as many as in the input)

    2021-11-17 Brian Maranville
    """
    output = copy(data)
    mappings = {
        "sf": {
            "+-": "-+",
            "-+": "+-"
        },
        "nsf": {
            "++": "--",
            "--": "++",
            "+": "-",
            "-": "+"
        }
    }
    def duplicate_and_remap_items(xs_type):
        mapping = mappings[xs_type]
        items = [d for d in data if d.polarization in mapping]
        for item in items:
            new_item = copy(item)
            new_item.polarization = mapping[item.polarization]
            output.append(new_item)

    if mix_sf:
        duplicate_and_remap_items("sf")

    if mix_nsf:
        duplicate_and_remap_items("nsf")

    return output

#@module
def align_background(data, align='auto'):
    """
    Determine the Qz value associated with the background measurement.

    The *align* flag determines which background points are matched
    to the sample points.  It can be 'sample' if background is
    measured using an offset from the sample angle, or 'detector'
    if it is offset from detector angle.   If *align* is 'auto', then
    use 'Qz_target' to align the background scan.

    For 'auto' alignment without Qz_target set, we can only distinguish
    relative and constant offsets, and cannot determine which of sample
    and detector is offset from the specular condition, so we must rely
    on convention. If the offset is constant for each angle, then it is
    assumed to be a sample offset. If the the offset is proportional to
    the angle (and therefore offset divided by angle is constant), then
    it is assumed to be a detector offset. If neither condition is met,
    it is assumed to be a sample offset.

    The 'auto' test is robust: 90% of the points should be within 5% of the
    median value of the vector for the offset to be considered a constant.

    **Inputs**

    data (refldata) : background data with unknown $q$

    align (opt:auto|sample|detector) : angle which determines $q_z$

    **Returns**

    output (refldata) : background with known $q$

    2015-12-17 Paul Kienzle
    2020-10-16 Paul Kienzle rename 'offset' to 'align'
    """
    from .background import set_background_alignment
    data = copy(data)
    set_background_alignment(data, align)
    return data


@module
def subtract_background(data, backp, backm, align="none"):
    """
    Subtract the background datasets from the specular dataset.

    The specular, background+ and background- signals should already be
    joined into single datasets. For each, the background is interpolated
    onto the specular Q values, extending above and below with the final
    background measurement. If there are no backgrounds, then data is
    sent through unchanged.

    Background subtraction is applied independently to the different
    polarization cross sections.

    The *align* flag determines which background points are matched
    to the sample points. It can be 'sample' if background is
    measured using an offset from the sample angle, or 'detector'
    if it is offset from detector angle. If it is 'none' then use
    the 'Qz_basis' value set in the loader. The 'auto' option uses
    'Qz_target' if it exists, or tries to guess from the measured angles.

    **Inputs**

    data (refldata) : specular data

    backp {Background+} (refldata?) : plus-offset background data

    backm {Background-} (refldata?) : minus-offset background data

    align (opt:none|sample|detector|auto) : apply align_background to
    background inputs with offset='auto'

    **Returns**

    output (refldata) : background subtracted specular data

    2016-03-23 Paul Kienzle
    """
    from .background import apply_background_subtraction

    # Note: This changes backp and backm, so copy first.
    if align != "none":
        if backp is not None:
            backp = copy(backp)
            align_background(backp, align=align)
        if backm is not None:
            backm = copy(backm)
            align_background(backm, align=align)

    #print "%s - (%s+%s)/2"%(data.name, (backp.name if backp else "none"), (backm.name if backm else "none"))
    data = copy(data)
    apply_background_subtraction(data, backp, backm)
    return data

@module
def interpolate_background(data, backp, backm, align='auto'):
    """
    Interpolate background data onto specular angles.

    The *align* flag determines which background points are matched
    to the specular points. The 'auto' option uses the Qz_basis set
    in the file loader, otherwise align the sample angle or the detector
    angle. This sets Qz_basis to align so that a subsequent subtraction
    operation will use the same interpolation.

    Masking of background or specular should occur before interpolation
    or after subtraction.

    **Inputs**

    data (refldata) : specular data

    backp {Background+} (refldata?) : plus-offset background data

    backm {Background-} (refldata?) : minus-offset background data

    align (opt:auto|sample|detector|sample_target|detector_target)
    : angle which determines $q_z$

    **Returns**

    output (refldata) : unchanged specular

    outp (refldata) : interpolated plus-offset background data

    outm (refldata) : interpolated minus-offset background data

    2020-11-20 Paul Kienzle new module
    """
    from .background import apply_interpolation

    if backp is not None:
        backp = copy(backp)
        backp.sample = copy(backp.sample)
        backp.detector = copy(backp.detector)
        apply_interpolation(data=backp, base=data, align=align)
    if backm is not None:
        backm = copy(backp)
        backm.sample = copy(backp.sample)
        backm.detector = copy(backp.detector)
        apply_interpolation(data=backm, base=data, align=align)
    return data, backp, backm

@module
def fit_background_field(back, fit_scale=True, scale=1.0, epsD0=0.01, epssi=1.109e-4, LS3=380, LS4=1269, LSD=1675, HD=150, maxF=76.2, Qcutoff=0.05):
    """
    Fit the background field from a thin liquid reservoir to background
    datasets. Background datasets:

     o Can be any at any (non-specular) condition

     o Should already be normalized by incident intensity

     o Can involve any number of scans

    The background datasets are fit using a Levenberg-Marquardt algorithm to a model
    involving a two parameters: epsD, the product of the incoherent attenuation coefficient
    of the reservoir (eps) and the reservoir thickness D, and a scale factor that accounts for
    uncertainty in the instrument geometry, i.e. the post-sample slit distances and/or the
    solid angle subtended by the detector.

    The uncertainty in the optimized parameters are estimated from the covariance matrix,
    and the chi-squared value of the fit (typically 1.5 or less) is also calculated and available
    in the parameter output. Note that the covariance matrix itself is passed to subract_background_field
    because the parameters are correlated.

    If the scale factor is not included in the fit, the accuracy of the calculation depends critically on
    correct values for the instrument geometry. The geometry is inferred from the data files;
    if it is missing, certain values can be overridden using the data entry boxes. These include the
    slit4-detector distance (0 for instruments with no slit4) and the detector height (for horizontal
    reflectometers, the width).

    The calculation assumes a geometry like that of the NIST liquid flow cell, in which the liquid reservoir
    is encased in silicon. For a different cell material (e.g. sapphire), the appropriate incoherent
    extinction coefficient for the material should replace the default Si value.

    **Inputs**

    back (refldata[]) : group of background datasets

    epsD0 {Reservoir extinction coefficient guess (mm^-1)} (float:mm^-1) : initial guess
    for product of incoherent extinction coefficient of reservoir and reservoir thickness

    epssi {Extinction coefficient of Si (mm^-1)} (float:mm^-1) : incoherent
    extinction coefficient of Si or cell materials

    fit_scale {Include scale factor in fit?} (bool) : True if scale factor on detector solid angle
    should be included in fit (use if uncertain of the instrument geometry)

    scale {Scale factor value} (float) : Value of scale factor. Initial guess if scale
    factor is included in fit; otherwise fixed scale factor value.

    Qcutoff {Target Qz cutoff (Ang^-1)} (float:Ang^-1) : Cutoff target Q_z value below which background data
    are excluded from background fit

    LS3 {sample-slit3 distance (mm)} (float:mm) : Distance from sample to slit3

    LS4 {sample-slit4 distance (mm)} (float:mm) : Distance from sample to slit4

    LSD {sample-detector distance (mm)} (float:mm) : Distance from sample to detector

    HD {detector height (mm)} (float:mm) : Height of detector

    maxF {maximum beam footprint (mm)} (float:mm) : Sample dimension in beam direction

    **Returns**

    fitparams (ncnr.refl.backgroundfield.params) : fit parameters, covariances, and chi-squared

    fit (refldata[]) : ReflData structure containing fit outputs (for plotting against
    background inputs to inspect background field fit)

    2018-06-12 David P. Hoogerheide; last updated 2018-10-12
    """

    from .backgroundfield import fit_background_field

    bff, outfitdata = fit_background_field(back, epsD0, epssi, fit_scale, scale, LS3, LS4, LSD, HD, Qcutoff, maxF)

    return bff, outfitdata


@module
def subtract_background_field(data, bfparams, epsD=None, epsD_var=None, scale=None, scale_var=None, scale_epsD_covar=None):
    """
    Subtract the background field from a thin liquid reservoir from
    a specular dataset, which should already be normalized by the incident intensity.
    Applies the background field fit with the "Fit Background Field" module. See that
    module's description for more details.

    **Inputs**

    data (refldata[]) : specular data

    bfparams (ncnr.refl.backgroundfield.params) : background field parameters

    epsD {epsD} (float) : p

    epsD_var {epsD variance} (float) : dp2

    scale {scale} (float) : s

    scale_var {scale factor variance} : ds2

    scale_epsD_covar {covariance of scale factor and epsD} : dsp

    **Returns**

    output (refldata[]) : background subtracted specular data

    2018-06-12 David P. Hoogerheide; last updated 2018-10-12
    """

    from .backgroundfield import apply_background_field_subtraction

    data = [copy(d) for d in data]

    if epsD is None: epsD = bfparams.p
    if epsD_var is None: epsD_var = bfparams.dp2
    if scale is None: scale = bfparams.s
    if scale_var is None: scale_var = bfparams.ds2
    if scale_epsD_covar is None: scale_epsD_covar = bfparams.dsp

    pcov = np.array([[scale_var, scale_epsD_covar],
                     [scale_epsD_covar, epsD_var]])

    apply_background_field_subtraction(data, epsD, bfparams.epssi, bfparams.LS3, bfparams.LS4, bfparams.L4D, bfparams.HD, bfparams.maxF, scale, pcov)

    return data

@module
def divide_intensity(data, base):
    """
    Scale data by incident intensity.

    Data is matched to the incident scan according to the measurement type:
    By default it is aligned by the angular resolution of both scans,
    assuming all data with the same angular resolution was subject to the
    same incident intensity.

    For Candor data it is aligned by the slit 1 opening (slit.x),
    and for MAGIK horizontal mode it is aligned by the incident angle (sample.angle_x)

    **Inputs**

    data (refldata) : specular, background or subtracted data

    base (refldata) : intensity data

    **Returns**

    output (refldata) : reflected intensity

    | 2015-12-17 Paul Kienzle
    | 2020-07-23 Brian Maranville add align_intensity flag to data to match incident and reflected points

    """
    if base is not None:
        from .scale import apply_intensity_norm
        data = copy(data)
        apply_intensity_norm(data, base)
    return data


@module
def smooth_slits(datasets, degree=1, span=2, dx=0.01):
    """
    Align slits with a moving window 1-D polynomial least squares filter.

    Updates *slit1.x*, *slit2.x* and *angular_resolution* attributes of the
    slit measurements so they all use a common set of points.

    Updates divergence automatically after smoothing.

    **Inputs**

    datasets (refldata[]) : slits to align and smooth

    degree (int) : polynomial degree on smoothing filter

    span (int) : number of consecutive points to use in the fit. Odd
    sized *span* is preferred.  *span* must be larger than *degree*.
    *degree=1* and *span=2* is equivalent to linear interpolation.

    dx (float:mm<0,>) :  size within which slits can be merged.

    **Returns**

    outputs (refldata[]) : aligned and smoothed slits.

    2015-12-17 Paul Kienzle
    """
    from .smoothslits import apply_smoothing
    datasets = [copy(d) for d in datasets]
    for d in datasets:
        d.slit1, d.slit2 = copy(d.slit1), copy(d.slit2)

    apply_smoothing(datasets, dx=dx, degree=degree, span=span)
    return datasets


@module
def abinitio_footprint(data, Io=1., width=None, offset=0.):
    """
    Apply an *ab initio* footprint correction to the data.

    Footprint is computed from the slits and the sample angle so those must
    be available in the data.  If the data has been stitched to common Q
    from different theta, lambda combinations, then footprint will no
    longer be available.

    Footprint is computed using sample angle. If background is measured using
    sample angle offset, then footprint should be applied before background
    subtraction. For detector angle offset the correction is the same for
    specular and background, so it can be applied before or after subtraction.

    **Inputs**

    data (refldata) : uncorrected measurement

    Io (float:): scale factor to account for vertical beam spill.

    width (float:mm) : sample width along the beam.  If not provided, use the
    value stored in the file.

    offset (float:mm) : offset of the center of rotation of the sample in
    the direction of the beam, toward the detector.

    **Returns**

    outputs (refldata): footprint-corrected data

    2016-09-02 Paul Kienzle
    """
    from .footprint import apply_abinitio_footprint

    data = copy(data)
    apply_abinitio_footprint(data, Io, width, offset)
    return data

@module
def fit_footprint(data, fit_range=[None, None], origin=False):
    """
    Fit a footprint using a range of data below the critical edge.

    If a range is not provided, then no footprint is fitted and instead the
    footprint slope and intercept from the *correct_footprint* component are
    used.

    **Inputs**

    data (refldata[]) : uncorrected measurement

    fit_range (range?:x): x-region over which to fit

    origin (bool) : True if data should go through the origin

    **Returns**

    fitted_footprint (ncnr.refl.footprint.params?) : slope and intercept

    2016-04-30 Paul Kienzle
    """
    from .footprint import fit_footprint
    if fit_range is None:
        fit_range = [None, None]
    footprint = fit_footprint(data, fit_range[0], fit_range[1], kind='slope' if origin else 'line')
    return footprint

@module
def correct_footprint(data, fitted_footprint, correction_range=[None, None],
                      slope=None, slope_error=0.0, intercept=None,
                      intercept_error=0.0):
    """
    Apply fitted footprint correction to each data set.

    If not footprint is fitted, then values must be entered for *slope* and
    *intercept*.

    **Inputs**

    data (refldata) : uncorrected measurement

    fitted_footprint (ncnr.refl.footprint.params?) : fitted footprint

    correction_range {Correction Application Range} (range?:x) : Lower bound of region to apply footprint correction

    slope (float) : footprint slope

    slope_error {Error on slope} (float): and uncertainty

    intercept (float) : footprint intercept

    intercept_error {Error on intercept} (float): and uncertainty

    **Returns**

    outputs (refldata): footprint-corrected data

    2017-06-29 Paul Kienzle
    """
    from .footprint import apply_fitted_footprint, FootprintData
    if correction_range is None:
        correction_range = [None, None]
    # always use manually-provided error on slope and intercept (not fitted)
    dp = np.array([slope_error, intercept_error])
    if fitted_footprint is None:
        # make empty footprint data object
        p = np.array([None, None])
        fitted_footprint = FootprintData(p, None)
    if slope is not None:
        fitted_footprint.p[0] = slope
    if intercept is not None:
        fitted_footprint.p[1] = intercept
    # in all cases, overwrite the error in fitted_footprint with specified
    # values:
    fitted_footprint.dp = dp
    data = copy(data)
    apply_fitted_footprint(data, fitted_footprint, correction_range)
    return data

@nocache
@module
def estimate_polarization(data, FRbalance=50.0, Emin=0.0, Imin=0.0, clip=False):
    """
    Compute polarizer and flipper efficiencies from the intensity data.

    If clip is true, reject points above or below particular efficiencies.
    The minimum intensity is 1e-10.  The minimum efficiency is 0.9.

    The computed values are systematically related to the efficiencies:

      beta: intensity/2
      fp: front polarizer efficiency is F
      rp: rear polarizer efficiency is R
      ff: front flipper efficiency is (1-x)/2
      rf: rear flipper efficiency is (1-y)/2

    reject is the indices of points which are clipped because they
    are below the minimum efficiency or intensity.

    See PolarizationEfficiency.pdf for details on the calculation.

    **Inputs**

    data (refldata[]) : direct beam measurement to determine polarization

    FRbalance (float:%<0,100>) : front/rear balance of to use for efficiency loss

    Emin (float:%<0,100>) : minimum efficiency cutoff

    Imin (float:counts/s<0,>) : minimum intensity cutoff

    clip {Clip efficiency} (bool) : clip efficiency between Emin and one

    **Returns**

    polarization (poldata) : estimated polarization correction factors

    2015-12-18 Paul Kienzle
    """
    from .polarization import PolarizationData

    poldata = PolarizationData(data, FRbal=0.01*FRbalance,
                               Emin=0.01*Emin, Imin=Imin, clip=clip)
    return poldata


@nocache
@module
def correct_polarization(data, polarization, spinflip=True):
    """
    Correct data for polarizer and flipper efficiencies.

    **Inputs**

    data (refldata[]) : polarized data to be corrected

    polarization (poldata) : estimated polarization efficiency

    spinflip {Correct spinflip data} (bool) : correct spinflip data if available

    **Returns**

    output (refldata[]) : polarization corrected data

    2015-12-18 Paul Kienzle
    | 2017-08-22 Brian Maranville interpolate back to Qz-basis for that cross-section
    """
    from .polarization import apply_polarization_correction
    data = copy(data)
    apply_polarization_correction(data, polarization, spinflip)
    return data


@module
def save(data, name='auto', ext='auto', path='auto'):
    """
    Save data to a particular file

    **Inputs**

    data (refldata) : data to save

    name (opt:auto|...) : name of the file, or 'auto' to use the basename

    ext {Extension} (opt:auto|...) : file extension, or 'auto' to use the
    id of the last step

    path (opt:auto|...) : data path, or 'auto' to use the current directory

    2015-12-17 Paul Kienzle
    """
    if path == 'auto':
        path = '.'
    if ext == 'auto':
        ext = '.dat'
    elif not ext.startswith('.'):
        ext = '.' + ext
    if name == 'auto':
        name = data.name
    filename = os.path.join(path, name+ext)
    data.save(filename)


@cache
@module
def ng7_psd(
        filelist=None,
        detector_correction=False,
        monitor_correction=False,
        center=None,
        intent='auto',
        sample_width=None,
        base='none'):
    r"""
    Load a list of NG7 PSD files from the NCNR data server.

    **Inputs**

    filelist (fileinfo[]): List of files to open.

    detector_correction {Apply detector deadtime correction} (bool)
    : If True, use deadtime constants in file to correct detector counts.

    monitor_correction {Apply monitor saturation correction} (bool)
    : If True, use the measured saturation curve in file to correct
    : the monitor counts.

    center {Beam center} (int)
    : Detector pixel containing the beam center.  This is needed for
    : plotting Qx-Qz, etc., and for setting the specular integration region.

    intent (opt:auto|specular|intensity|scan)
    : Measurement intent (specular, slit, or some other scan), auto or infer.
    : If intent is 'scan', then use the first scanned variable.

    sample_width {Sample width (mm)} (float?)
    : Width of the sample along the beam direction in mm, used for
    calculating the effective resolution when the sample is smaller
    than the beam.  Leave blank to use value from data file.

    base {Normalize by} (opt:auto|monitor|time|roi|power|none)
    : How to convert from counts to count rates.
    : Leave this as none if your template does normalization after integration.

    **Returns**

    output (refldata[]): All entries of all files in the list.

    | 2020-02-05 Paul Kienzle
    | 2020-02-11 Paul Kienzle include divergence estimate in startup
    """
    from .load import url_load_list
    from .ng7psd import load_entries

    # Note: divergence is required for join, so always calculate it.  If you
    # really want it optional then use:
    #
    #  auto_divergence {Calculate dQ} (bool)
    #    : Automatically calculate the angular divergence of the beam.
    #
    auto_divergence = True

    datasets = []
    for data in url_load_list(filelist, loader=load_entries):
        data.Qz_basis = 'target'
        if intent not in [None, 'auto']:
            data.intent = intent
        if center is not None:
            data = psd_center(data, center)
        if auto_divergence:
            data = divergence(data, sample_width)
        if detector_correction:
            data = detector_dead_time(data, None)
        if monitor_correction:
            data = monitor_saturation(data)
        data = normalize(data, base=base)
        #print "data loaded and normalized"
        datasets.append(data)

    return datasets

@cache
@module
def super_load(filelist=None,
               detector_correction=False,
               monitor_correction=False,
               intent='auto',
               Qz_basis='actual',
               sample_width=None,
               base='auto'):
    r"""
    Load a list of nexus files from the NCNR data server.

    *Qz_basis* uses one of the following values:

        **actual**
            calculates Qx and Qz as (x,z)-components of
            $(\vec k_{\text{out}} - \vec k_\text{in})$ in sample coordinates,
        **detector**
            ignores the sample angle and calculates Qz
            as $(4\pi/\lambda \sin(\theta_\text{detector}/2))$,
        **sample**
            ignores the detector angle and calculates Qz
            as $(4\pi/\lambda \sin(\theta_\text{sample}))$
        **target**
            uses the user-supplied Qz_target values

    **Inputs**

    filelist (fileinfo[]): List of files to open.

    detector_correction {Apply detector deadtime correction} (bool)
    : Which deadtime constant to use for detector deadtime.

    monitor_correction {Apply monitor deadtime correction} (bool)
    : Which deadtime constant to use for monitor deadtime.

    intent (opt:auto|specular|background+\|background-\|intensity|rock sample|rock detector|rock qx|scan)
    : Measurement intent (specular, background+, background-, slit, rock),
    auto or infer.  If intent is 'scan', then use the first scanned variable.

    Qz_basis (opt:actual|detector|sample|target)
    : How to calculate Qz from instrument angles.

    sample_width {Sample width (mm)} (float?)
    : Width of the sample along the beam direction in mm, used for
    calculating the effective resolution when the sample is smaller
    than the beam.  Leave blank to use value from data file.

    base {Normalize by} (opt:auto|monitor|time|roi|power|none)
    : how to convert from counts to count rates

    **Returns**

    output (refldata[]): All entries of all files in the list.

    | 2017-01-13 Brian Maranville
    | 2017-02-15 Paul Kienzle normalize by time if monitor is not present
    | 2017-08-21 Brian Maranville use fileName from trajectory
    | 2018-05-01 Brian Maranville import temperature metadata
    | 2018-05-07 Brian Maranville detector deadtime correction defaults to True
    | 2018-05-10 Brian Maranville export all columns if intent is scan
    | 2018-05-11 Brian Maranville detector deadtime correction defaults to False
    | 2018-06-18 Brian Maranville change to nexusref to ignore areaDetector
    | 2018-06-20 Brian Maranville promote detector.wavelength to column (and resolution)
    | 2018-08-29 Paul Kienzle ignore sampleTilt field for NG7
    | 2018-12-10 Brian Maranville get_plottable routines moved to python data container from js
    | 2020-01-21 Brian Maranville updated loader to handle hdf-nexus
    | 2020-12-18 Brian Maranville adding source_power column to monitor
    | 2022-04-27 Brian Maranville fix count time for Rigaku in sweep mode
    """
    from .load import url_load_list
    #from .intent import apply_intent
    #from .angles import apply_divergence
    # Note: Fileinfo is a structure with
    #     { path: "location/on/server", mtime: timestamp }

    # Note: divergence is required for join, so always calculate it.  If you
    # really want it optional then use:
    #
    #  auto_divergence {Calculate dQ} (bool)
    #    : Automatically calculate the angular divergence of the beam.
    #
    auto_divergence = True

    # TODO: sample_width is ignored if datafile defines angular_divergence

    datasets = []
    for data in url_load_list(filelist):
        data.Qz_basis = Qz_basis
        if intent not in [None, 'auto']:
            data.intent = intent
        if auto_divergence:
            data = divergence(data, sample_width)
        if detector_correction:
            data = detector_dead_time(data, None)
        if monitor_correction:
            data = monitor_dead_time(data, None)
        data = normalize(data, base=base)
        #print "data loaded and normalized"
        datasets.append(data)

    return datasets

@cache
@module
def magik_horizontal(filelist=None,
                    detector_correction=False,
                    monitor_correction=False,
                    intent='auto',
                    Qz_basis='actual',
                    sample_width=None,
                    base='auto'):
    r"""
    Load a list of nexus files from the NCNR data server.

    *Qz_basis* uses one of the following values:

        **actual**
            calculates Qx and Qz as (x,z)-components of
            $(\vec k_{\text{out}} - \vec k_\text{in})$ in sample coordinates,
        **detector**
            ignores the sample angle and calculates Qz
            as $(4\pi/\lambda \sin(\theta_\text{detector}/2))$,
        **sample**
            ignores the detector angle and calculates Qz
            as $(4\pi/\lambda \sin(\theta_\text{sample}))$
        **target**
            uses the user-supplied Qz_target values

    **Inputs**

    filelist (fileinfo[]): List of files to open.

    detector_correction {Apply detector deadtime correction} (bool)
    : Which deadtime constant to use for detector deadtime.

    monitor_correction {Apply monitor deadtime correction} (bool)
    : Which deadtime constant to use for monitor deadtime.

    intent (opt:auto|specular|background+\|background-\|intensity|rock sample|rock detector|rock qx|scan)
    : Measurement intent (specular, background+, background-, slit, rock),
    auto or infer.  If intent is 'scan', then use the first scanned variable.

    Qz_basis (opt:actual|detector|sample|target)
    : How to calculate Qz from instrument angles.

    sample_width {Sample width (mm)} (float?)
    : Width of the sample along the beam direction in mm, used for
    calculating the effective resolution when the sample is smaller
    than the beam.  Leave blank to use value from data file.

    base {Normalize by} (opt:auto|monitor|time|roi|power|none)
    : how to convert from counts to count rates

    **Returns**

    output (refldata[]): All entries of all files in the list.

    | 2020-07-21 Brian Maranville
    | 2020-07-23 Brian Maranville Added a flag to the loader, to control divide_intensity align_by
    | 2020-09-03 Brian Maranville Vertical slit readout changed
    | 2021-09-20 Brian Maranville use horizontalGeom.angle for sample.angle_x (ignore tilt except in ROCK)
    """
    from .load import url_load_list
    from .magik_horizontal import load_entries
    
    # Note: Fileinfo is a structure with
    #     { path: "location/on/server", mtime: timestamp }

    # Note: divergence is required for join, so always calculate it.  If you
    # really want it optional then use:
    #
    #  auto_divergence {Calculate dQ} (bool)
    #    : Automatically calculate the angular divergence of the beam.
    #
    auto_divergence = True

    # TODO: sample_width is ignored if datafile defines angular_divergence

    datasets = []
    for data in url_load_list(filelist, loader=load_entries):
        data.Qz_basis = Qz_basis
        if intent not in [None, 'auto']:
            data.intent = intent
        if auto_divergence:
            data = divergence(data, sample_width)
        if detector_correction:
            data = detector_dead_time(data, None)
        if monitor_correction:
            data = monitor_dead_time(data, None)
        data = normalize(data, base=base)
        #print "data loaded and normalized"
        datasets.append(data)

    return datasets

@cache
@module
def super_load_sorted(filelist=None,
                      detector_correction=False,
                      monitor_correction=False,
                      sample_width=None,
                      base='auto'):
    """
    Load a list of nexus files from the NCNR data server, to be sorted by
    the intent stored in the file.  If intent does not match
    'specular', 'background+', 'background-' or 'intensity', it is not returned.

    **Inputs**

    filelist (fileinfo[]): List of files to open.

    detector_correction {Apply detector deadtime correction} (bool)
    : Which deadtime constant to use for detector deadtime.

    monitor_correction {Apply monitor deadtime correction} (bool)
    : Which deadtime constant to use for monitor deadtime.

    sample_width {Sample width (mm)} (float): Width of the sample along the
    beam direction in mm, used for calculating the effective resolution when
    the sample is smaller than the beam.  Leave blank to use value from data file.

    base {Normalize by} (opt:auto|monitor|time|roi|power|none)
    : how to convert from counts to count rates

    **Returns**

    spec (refldata[]): All entries of all spec files in the list.

    bgp (refldata[]): All entries of all bg+ files in the list.

    bgm (refldata[]): All entries of all bg- files in the list.

    slit (refldata[]): All entries of all slit files in the list.

    2016-06-30 Brian Maranville
    """
    from .load import url_load_list
    auto_divergence = True
    sorting_key = "intent"
    sort_values = ["specular", "background+", "background-", "intensity"]
    outputs = dict([(key, []) for key in sort_values])

    for data in url_load_list(filelist):
        if auto_divergence:
            data = divergence(data, sample_width)
        if detector_correction:
            data = detector_dead_time(data, None)
        if monitor_correction:
            data = monitor_dead_time(data, None)
        data = normalize(data, base=base)
        intent = getattr(data, sorting_key, None)
        if intent in outputs:
            outputs[intent].append(data)

    return tuple([outputs[k] for k in sort_values])

@module
def spin_asymmetry(data):
    """
    Do the calculation (up-up - down-down) / (up-up + down-down) and
    return a single dataset.

    **Inputs**

    data (refldata[]): input data; must contain up-up and down-down polarizations

    **Returns**

    output (refldata): calculated spin asymmetry.

    2016-04-04 Brian Maranville
    """
    mm = [d for d in data if d.polarization in ('-', '--')][0]
    pp = [d for d in data if d.polarization in ('+', '++')][0]
    output = copy(mm)
    output.vscale = "linear"
    output.vlabel = "Spin asymmetry (pp-mm)/(pp+mm) "
    output.vunits = "unitless"
    shortest = min(mm.v.shape[0], pp.v.shape[0])
    mmv = mm.v[:shortest]
    mmdv = mm.dv[:shortest]
    ppv = pp.v[:shortest]
    ppdv = pp.dv[:shortest]
    denom = (mmv + ppv)
    output.v = (ppv - mmv) / denom
    # d(sa)/d(x) = 2*x/(x+y)**2, d(sa)/d(y) = -2*y/(x+y)**2
    output.dv = np.sqrt(((2.0*mmv*mmdv)/(denom**2))**2 + ((2.0*ppv*ppdv)/(denom**2))**2)
    return output

@module
def average_flux(data, base, beam_height=25):
    """
    Calculate time-averaged flux on the sample

    Data is matched according to angular resolution, assuming all data with
    the same angular resolution was subject to the same incident intensity.
    Does not work on polarized beam data with multiple slit scans

    Beam area is taken to be beam_height * slit2 aperture (from file)

    **Inputs**

    data (refldata[]) : specular, background or subtracted data

    base (refldata) : intensity data

    beam_height (float:mm): height of the beam at the sample position

    **Returns**

    flux (ncnr.refl.flux.params?) : integrated flux data

    2018-03-01 Brian Maranville
    """
    from reductus.dataflow.modules import refl
    TIME_RESOLUTION = 1e-6 # 1 microsecond for NCNR timers.

    if base is not None:
        from .scale import calculate_number
        from reductus.dataflow.lib import err1d

        fluxes = []
        total_number = 0.0
        total_number_variance = 0.0
        total_time = 0.0
        total_time_variance = 0.0
        sum_weighted_flux = 0.0
        sum_weighted_flux_variance = 0.0

        for datum in data:
            datum = copy(datum)
            beam_area = datum.slit2.x * beam_height / 100.0 # both in mm, convert to cm
            N, varN = calculate_number(datum, base, time_uncertainty=TIME_RESOLUTION)
            S, varS = err1d.sum(N, varN)
            P, varP = err1d.div(N, varN, beam_area, 0.0)
            A, varA = err1d.sum(P, varP) # time-weighted average of Flux/Area
            T, varT = err1d.sum(datum.monitor.count_time, TIME_RESOLUTION**2)
            F, varF = err1d.div(A, varA, T, varT) # average Flux/(Area * Time)
            fluxes.append({
                "name": datum.name,
                "number_incident": S,
                "number_incident_error": np.sqrt(varS),
                "number_incident_units": "neutrons",
                "average_flux": F,
                "average_flux_error": np.sqrt(varF),
                "average_flux_units": "neutrons/(second * cm^2)",
                "total_time": float(T),
                "total_time_error": float(np.sqrt(varT))
            })
            total_number += S
            total_number_variance += varS
            total_time += T
            total_time_variance += varT
            sum_weighted_flux += A
            sum_weighted_flux_variance += varA
        aggregated_flux, aggregated_flux_variance = err1d.div(sum_weighted_flux, sum_weighted_flux_variance, total_time, total_time_variance)
        output = refl.FluxData(fluxes, {
            "aggregated_average_flux": aggregated_flux,
            "aggregated_average_flux_error": np.sqrt(aggregated_flux_variance),
            "aggregated_time": total_time,
            "aggregated_time_error": np.sqrt(total_time_variance)
        })
    else:
        output = refl.FluxData([], None)
    return output



"""
Add an annotation to an exception giving a context for the embedded exception.

For example::

    try:
        operate_on_file(filename)
    except:
        annotate_exception("while operating on "+filename)
        raise

This is not so useful in python 3.5, with its ability to create a chained
exception using *raise Exception() from exc*.
"""
import sys

def annotate_exception(msg, exc=None):
    """
    Add an annotation to the current exception, which can then be forwarded
    to the caller using a bare "raise" statement to reraise the annotated
    exception.
    """
    if not exc:
        exc = sys.exc_info()[1]

    args = exc.args
    if isinstance(exc, (OSError, IOError)) and len(args) == 2:
        # Special handling of system errors with args=(errno, message)
        exc.args = (args[0], " ".join((args[1], msg)))
    elif not args:
        exc.args = (msg,)
    else:
        exc.args = tuple([" ".join((args[0], msg))] + list(args[1:]))



# -*- coding: UTF-8 -*-
from collections import OrderedDict
import datetime
import numpy as np
import json
import io
import sys

IS_PY3 = sys.version_info[0] >= 3

from reductus.dataflow.lib.uncertainty import Uncertainty
from reductus.dataflow.lib.strings import _s, _b
from reductus.dataflow.lib.exporters import exports_text, exports_HDF5

class RawData(object):
    def __init__(self, metadata=None, countTime=None, detCts=None, transCts=None, monCts=None, Q=None):
        self.metadata = metadata
        self.countTime = countTime
        self.detCts = Uncertainty(detCts, detCts + (detCts == 0))
        self.transCts = Uncertainty(transCts, transCts + (transCts == 0))
        self.monCts = Uncertainty(monCts, monCts + (monCts == 0))
        self.Q = Q
        self.Q_offset = 0.0

    def todict(self):
        return _toDictItem(self.metadata)

    def get_plottable(self):
        return {"entry": "entry", "type": "metadata", "values": _toDictItem(self.metadata)}

    def get_metadata(self):
        return _toDictItem(self.metadata)

class USansData(RawData):
    xcol = "Q"
    xlabel = "Q"
    xunits = "1/Ang"
    ycol = "I_det"
    ylabel = "Counts"
    yunits = "arb. units"
    def get_plottable(self):
        name = self.metadata.get("run.filename", "")
        entry = ""
        xcol = "Q"
        ycol = "I_det"
        columns = OrderedDict([
            ('Q', {'label': 'Q', 'units': '1/Ang'}),
            ('I_det', {'label': 'Counts', 'units': '', 'errorbars': 'dI'}),
            ('I_trans', {'label': 'Transmission Counts', 'units': '', 'errorbars': 'dIt'}),
            ('I_mon', {'label': 'Monitor Counts', 'units': '', 'errorbars': 'dIm'}),
            ('countTime', {'label': 'Count Time', 'units': 's'}),
        ])
        datas = {
            "Q": {"values": self.Q.tolist()},
            "I_det": {"values": self.detCts.x.tolist(), "errorbars": np.sqrt(self.detCts.variance).tolist()},
            "I_trans": {"values": self.transCts.x.tolist(), "errorbars": np.sqrt(self.transCts.variance).tolist()},
            "I_mon": {"values": self.monCts.x.tolist(), "errorbars": np.sqrt(self.monCts.variance).tolist()},
            "countTime": {"values": self.countTime.tolist()},
        }
        series = [{"label": "%s:%s" % (name, entry)}]
        plottable = {
            "type": "nd",
            "title": "%s:%s" % (name, entry),
            "entry": "entry",
            "columns": columns,
            "options": {
                "series": series,
                "axes": {
                    "xaxis": {"label": "%s(%s)" % (columns[xcol]["label"], columns[xcol]["units"])},
                    "yaxis": {"label": "%s(%s)" % (columns[ycol]["label"], columns[ycol]["units"])}
                },
                "xcol": xcol,
                "ycol": ycol,
                "errorbar_width": 0
            },
            "datas": datas
        }
        #print(plottable)
        return plottable

class USansCorData(object):
    def __init__(self, metadata=None, iqCOR=None, Q=None):
        self.metadata = metadata
        self.iqCOR = iqCOR
        self.Q = Q

    def todict(self):
        return _toDictItem(self.metadata)

    def get_plottable(self):
        if isinstance(self.metadata, list):
            name = ",".join([m.get("Sample file", "") for m in self.metadata])
        else:
            name = self.metadata.get("Sample file", "")
        entry = ""
        xcol = "Q"
        ycol = "iqCOR"
        columns = OrderedDict([
            ('Q', {'label': 'Q', 'units': '1/Ang'}),
            ('iqCOR', {'label': 'Corrected Counts', 'units': '/1E6 Monitor', 'errorbars': 'dI'}),
        ])
        datas = {
            "Q": {"values": self.Q.tolist()},
            "iqCOR": {"values": self.iqCOR.x.tolist(), "errorbars": np.sqrt(self.iqCOR.variance).tolist()},
        }
        series = [{"label": "%s:%s" % (name, entry)}]
        plottable = {
            "type": "nd",
            "title": "%s:%s" % (name, entry),
            "entry": "entry",
            "columns": columns,
            "options": {
                "series": series,
                "axes": {
                    "xaxis": {"label": "%s(%s)" % (columns[xcol]["label"], columns[xcol]["units"])},
                    "yaxis": {"label": "%s(%s)" % (columns[ycol]["label"], columns[ycol]["units"])}
                },
                "xcol": xcol,
                "ycol": ycol,
                "errorbar_width": 0
            },
            "datas": datas
        }
        #print(plottable)
        return plottable

    def get_metadata(self):
        return _toDictItem(self.metadata)
    
    @exports_text(name="column")
    def to_column_text(self):
        from io import BytesIO
        # export to 6-column format compatible with SASVIEW
        # Data columns are Qx - Qy - I(Qx,Qy) - err(I) - Qz - SigmaQ_parall - SigmaQ_perp - fSubS(beam stop shadow)
        labels = ["Qx (1/A)", "I(Q) (Counts/sec/(1e6 Monitor))", "std. dev. I(Q) (1/cm)", "dQ (1/A)", "filler", "filler"]

        with BytesIO() as fid:
            cleaned_metadata = _toDictItem(self.metadata, convert_bytes=True)
            if not isinstance(cleaned_metadata, list):
                cleaned_metadata = [cleaned_metadata]
            for c in cleaned_metadata:
                fid.write(b'### Parameters:\n')
                for k,v in c.items():
                    fid.write(_b("# %s\n" % json.dumps({k: v}).strip("{}")))
            #fid.write(_b("# %s\n" % json.dumps(_toDictItem(self.metadata, convert_bytes=True)).strip("{}")))
            fid.write(_b("# %s\n" % json.dumps({"columns": labels}).strip("{}")))
            filler = np.ones_like(self.Q, dtype='float') * -1.0 * cleaned_metadata[0]["dQv"]
            np.savetxt(fid, np.vstack([self.Q, self.iqCOR.x, np.sqrt(self.iqCOR.variance), filler, filler, filler]).T, fmt="%15.6g")
            fid.seek(0)
            name = _s(cleaned_metadata[0]["Sample file"])
            entry = ""
            value = fid.read()

        return {
            "name": name,
            "entry": "",
            "file_suffix": ".usans.cor",
            "value": value.decode(),
        }
    
    @exports_HDF5(name="NXcanSAS")
    def to_NXcanSAS(self):
        import h5py
        from io import BytesIO

        fid = BytesIO()
        h5_item = h5py.File(fid, 'w')
        string_dt = h5py.string_dtype(encoding='utf-8')

        metadata = self.metadata[0]
        #entry_name = metadata.get("entry", "entry")
        nxentry = h5_item.create_group("sasentry1")
        nxentry.attrs.update({
            "NX_class": "NXentry",
            "canSAS_class": "SASentry",
            "version": "1.0"
        })
        nxentry["definition"] = "NXcanSAS"
        nxentry["run"] = _s(metadata.get("Sample file", "default_name"))
        nxentry["title"] = ""

        datagroup = nxentry.create_group("sasdata")
        datagroup.attrs.update({
            "NX_class": "NXdata",
            "canSAS_class": "SASdata",
            "signal": "I",
            "I_axes": "Q",
            "Q_indices": 0,
            "timestamp": _s(metadata.get("Start time", "unknown")),
        })
        datagroup["I"] = self.iqCOR.x
        datagroup["I"].attrs["units"] = "1/cm"
        datagroup["I"].attrs["uncertainties"] = "Idev"
        datagroup["Idev"] = np.sqrt(self.iqCOR.variance)
        datagroup["Idev"].attrs["units"] = "1/cm"
        datagroup["Q"] = self.Q
        datagroup["Q"].attrs["units"] = "1/angstrom"
        datagroup["Q"].attrs["resolutions"] = "dQl,dQw"
        datagroup["dQl"] = np.ones_like(self.Q, dtype='float') * -1.0 * metadata.get("dQv", 0.0)
        datagroup["dQl"].attrs["units"] = "1/angstrom"
        datagroup["dQw"] = np.zeros_like(self.Q, dtype='float')
        datagroup["dQw"].attrs["units"] = "1/angstrom"

        instrument = nxentry.create_group("sasinstrument")
        instrument.attrs.update({
            "canSAS_class": "SASinstrument",
            "NX_class": "NXinstrument"
        })
        sasaperture = instrument.create_group("sasaperture")
        sasaperture.attrs.update({
            "canSAS_class": "SASaperture",
            "NX_class": "NXaperture"
        })
        sasaperture["shape"] = "slit"
        sasaperture["x_gap"] = np.array([0.1], dtype='float')
        sasaperture["x_gap"].attrs["units"] = "cm"
        sasaperture["y_gap"] = np.array([5.0], dtype='float')
        sasaperture["y_gap"].attrs["units"] = "cm"
        sasdetector = instrument.create_group("sasdetector")
        sasdetector.attrs.update({
            "canSAS_class": "SASdetector",
            "NX_class": "NXdetector"
        })
        sasdetector["name"] = "BT5 DETECTOR ARRAY"
        sassource = instrument.create_group("sassource")
        sassource.attrs.update({
            "canSAS_class": "sassource",
            "NX_class": "NXdetector"
        })
        sassource["incident_wavelength"] = np.array([2.38], dtype='float')
        sassource["incident_wavelength"].attrs["units"] = "A"
        sassource["incident_wavelength_spread"] = np.array([0.06], dtype='float')
        sassource["incident_wavelength_spread"].attrs["units"] = "A"
        sassource["radiation"] = "Reactor Neutron Source"

        sasprocess = nxentry.create_group("sasprocess")
        sasprocess.attrs.update({
            "canSAS_class": "SASprocess",
            "NX_class": "NXprocess"
        })
        sasprocess["name"] = "NIST reductus"

        sassample = nxentry.create_group("sassample")
        sassample.attrs.update({
            "canSAS_class": "SASsample",
            "NX_class": "NXsample"
        })
        sassample["name"] = ""
        sassample["thickness"] = np.array([metadata.get("Thickness")], dtype='float')
        sassample["thickness"].attrs["units"] = "cm"
        sassample["transmission"] = np.array([metadata.get("Trock/Twide")], dtype='float')
        sassample["transmission"].attrs["units"] = "unitless"


        return {
            "name": _s(metadata.get("Sample file", "default_name")),
            "entry": _s(metadata.get("entry", "default_entry")),
            "file_suffix": ".usans.cor.h5",
            "value": h5_item,
        }

class Parameters(dict):
    def get_metadata(self):
        return _toDictItem(self)

    def get_plottable(self):
        return {"entry": "entry", "type": "metadata", "values": _toDictItem(self)}


def _toDictItem(obj, convert_bytes=False):
    if isinstance(obj, np.integer):
        obj = int(obj)
    elif isinstance(obj, np.floating):
        obj = float(obj)
    elif isinstance(obj, np.ndarray):
        obj = obj.tolist()
    elif isinstance(obj, datetime.datetime):
        obj = [obj.year, obj.month, obj.day, obj.hour, obj.minute, obj.second]
    elif isinstance(obj, list):
        obj = [_toDictItem(a, convert_bytes=convert_bytes) for a in obj]
    elif isinstance(obj, dict):
        obj = OrderedDict([(k, _toDictItem(v, convert_bytes=convert_bytes)) for k, v in obj.items()])
    elif isinstance(obj, bytes) and convert_bytes == True:
        obj = obj.decode()
    return obj



"""
Calculations can be cached, either using a redis server or using a local
in-memory cache.  The redis server is useful when serving calculations
over the net via CGI since it can be used by different processes.

A singleton :class:`CacheManager` is available for programs that only
need a single shared cache.  Call *cache.use_redis(redis args)* during
program configuration to set up redis, otherwise the default is to use
an in-memory cache.   The calculation library will call *cache.get_cache()*
to retrieve the cache connection, allowing calculations to be memoized.
"""
import warnings
import sys
import os
import subprocess
import time
import tempfile

try:
    # CRUFT: use cPickle for python 2.7
    import cPickle as pickle
except ImportError:
    import pickle

PICKLE_PROTOCOL = pickle.HIGHEST_PROTOCOL # use the best

def memory_cache():
    from . import fakeredis
    return fakeredis.MemoryCache()

def file_cache(cachedir="~/.reductus/cache"):
    from . import fakeredis
    return fakeredis.FileBasedCache(cachedir=cachedir)


# port 6379 is the default port value for the python redis connection
def redis_connect(host="localhost", port=6379, maxmemory=4.0, **kwargs):
    """
    Open a redis connection.

    If host is localhost, then try starting the redis server.

    If redis is unavailable, then return a simple dict cache.
    """
    import redis  # lazy import so that redis need not be available

    # ensure redis is running, at least if we are not on a windows box
    try:
        cache = redis.Redis(host=host, port=port, **kwargs)
        # first, check to see if it is already running:
        cache.ping()
    except redis.exceptions.ConnectionError:
        # if it's not running, and this is a platform on which we can start it:
        if host == "localhost" and not sys.platform == 'win32':
            subprocess.Popen(["redis-server"],
                             stdout=open(os.devnull, "w"),
                             stderr=subprocess.STDOUT)
            time.sleep(10)
            cache = redis.Redis(host=host, port=port, **kwargs)
            cache.ping()
        else:
            raise

    # set the memory settings for already-running Redis:
    cache.config_set("maxmemory", "%d" % (int(maxmemory*2**30),))
    cache.config_set("maxmemory-policy", "allkeys-lru")

    return cache

class CacheManager(object):
    """
    Manage the connection to the key-value cache.
    """
    def __init__(self):
        self._cache = None
        self._file_cache = None
        self._engine = None
        self._use_compression = False
        self._pickle_protocol = PICKLE_PROTOCOL

    @property
    def engine(self):
        return self._engine

    def use_memory(self):
        """
        Set up cache for testing.
        """
        if self._cache is None:
            cachedir = os.path.join(tempfile.gettempdir(), "reductus_test")
            self._cache = memory_cache()
            self._file_cache = file_cache(cachedir=cachedir)
            self._cache_engine = "memory"

    def use_diskcache(self, **kwargs):
        """
        use the PyPi package 'diskcache' as the main store
        """
        try:
            from diskcache import FanoutCache as Cache
            # patch the class so it has "exists" method
            Cache.exists = Cache.__contains__
            cachedir = kwargs.pop("cachedir", "cache")
            self._cache = Cache(cachedir, **kwargs)
            file_cachedir = kwargs.pop("file_cachedir", "files_cache")
            self._file_cache = Cache(file_cachedir, **kwargs)
            self._cache_engine = "diskcache"
            return
        except Exception as exc:
            warning = "diskcache connection failed with:\n\t" + str(exc)
            warning += "\nFalling back to in-memory cache."
            warnings.warn(warning)
            self.use_memory()
        

    def use_redis(self, **kwargs):
        """
        Use redis for managing the cache.

        The arguments given to use_redis will be used to connect to the
        redis server when get_cache() is called.   See *redis.Redis()* for
        details.  If use_redis() is not called, then get_cache() will use
        an in-memory cache instead.
        """
        try:
            self._cache = redis_connect(**kwargs)
            self._file_cache = self._cache
            self._cache_engine = "redis"
            return
        except Exception as exc:
            warning = "Redis connection failed with:\n\t" + str(exc)
            warning += "\nFalling back to in-memory cache."
            warnings.warn(warning)
            self.use_memory()

    def get_cache(self):
        """
        Connect to the key-value cache.
        """
        if self._cache is None:
            self.use_memory()
        return self._cache

    def get_cache_manager(self):
        """
        return this manager class
        """
        if self._cache is None:
            self.use_memory()
        return self

    def get_file_cache(self):
        """
        Connect to the file cache.
        """
        if self._cache is None:
            self.use_memory()
        return self._file_cache

    def store_file(self, key, contents):
        if self._use_compression:
            import lz4.frame
            contents = lz4.frame.compress(contents)
        self._file_cache.set(key, contents)

    def retrieve_file(self, key):
        contents = self._file_cache.get(key)
        if self._use_compression:
            import lz4.frame
            contents = lz4.frame.decompress(contents)
        return contents

    def store(self, key, value):
        string = pickle.dumps(value, protocol=self._pickle_protocol)
        if self._use_compression:
            import lz4.frame
            string = lz4.frame.compress(string)
        self._cache.set(key, string)

    def retrieve(self, key):
        string = self._cache.get(key)
        if self._use_compression:
            import lz4.frame
            string = lz4.frame.decompress(string)
        value = pickle.loads(string)
        return value

    def delete(self, key):
        self._cache.delete(key)

    def file_exists(self, key):
        return self._file_cache.exists(key)
        
    def exists(self, key):
        return self._cache.exists(key)


# Singleton cache manager if you only need one cache
CACHE_MANAGER = CacheManager()

# direct access to singleton methods
use_redis = CACHE_MANAGER.use_redis
use_diskcache = CACHE_MANAGER.use_diskcache
get_cache = CACHE_MANAGER.get_cache_manager
get_file_cache = CACHE_MANAGER.get_file_cache
set_test_cache = CACHE_MANAGER.use_memory



"""
File loaders for NCNR VAX SANS format
"""
from __future__ import print_function
import sys
import os
import struct
import time
import math

try:
    # pylint: disable=unused-import
    from typing import Tuple, Dict, List, Optional, Sequence, Any, Iterator, AnyStr
    # pylint: enable=unused-import
    ignore = 0  # keep pycharm happy
except ImportError:
    pass

import numpy as np

# CRUFT: python 2.x needs to convert unicode to str; 3.x leaves it as unicode
if sys.version_info[0] >= 3:
    def bytes_to_str(s):
        # type: (AnyStr) -> str
        return s.decode('utf-8') if isinstance(s, bytes) else s # type: ignore
    def str_to_bytes(s):
        # type: (AnyStr) -> bytes
        """Return byte string or list of byte strings"""
        return s.encode('utf-8') if isinstance(s, str) else s # type: ignore
    def write(fid, s):
        # type: (IO[Any], AnyStr) -> None
        #print("fid type=%s"%type(fid))
        fid.write(str_to_bytes(s))
else: # python 2.x
    def bytes_to_str(s):
        # type: (AnyStr) -> str
        return s
    def str_to_bytes(s):
        # type: (AnyStr) -> bytes
        """Return byte string or list of byte strings"""
        return s
    def write(fid, s):
        # type: (IO[Any], AnyStr) -> None
        fid.write(s)

__all__ = ["load", "save"]

DATETIME_FORMAT = '%d-%b-%Y %H:%M:%S'
DATE_FORMAT = '%d%b%y '

class SansFormatInfo:
    #: type for each field
    types = None  # type: Dict[str, str]
    #: units for each field
    units = None  # type: Dict[str, str]
    #: field names
    fields = None  # type: List[str]
    #: integer fields
    ints = None  # type: List[str]
    #: float fields
    reals = None  # type: List[str]
    #: string fields
    strings = None  # type: List[str]
    #: struct pack/unpack codes for header
    header_struct = None  # type: str
    #: struct pack/unpack codes for data
    data_struct = None  # type: str
    #: precision for each field
    precision = None  # type: Dict[str, float]
    def __init__(self, *args):
        # type: (*Tuple[str, str]) -> None
        self.types = dict(args)
        self.fields = [name for name, _ in args]
        dtype_to_struct_type = {
            'vaxR4': '4s',
            DATETIME_FORMAT: '20s',
            DATE_FORMAT: '8s',
            }
        self.header_struct = "<"+"".join(dtype_to_struct_type.get(dtype, dtype)
                                         for _, dtype in args)
        assert struct.calcsize(self.header_struct) == 512
        self.ints = [name for name, dtype in args if dtype == 'i']
        self.reals = [name for name, dtype in args if dtype == 'vaxR4']
        self.strings = [name for name, dtype in args if dtype.endswith('s')]
        self.data_struct = '<16401h'
        self.units = dict([
            ('run.ctime', 's'),           #  27  count time per prefactor (s)
            ('run.rtime', 's'),           #  31  total count time (s)
            ('sample.thk', 'cm'),         # 162  sample thickness (cm)
            ('sample.rotang', 'cm'),      # 170  sample rotation angle (degrees)
            ('sample.temp', 'C'),         # 186  sample temperature (C)
            ('det.calx1', 'mm'),          # 220  detector x pixel size (mm)
            ('det.caly1', 'mm'),          # 232  detector y pixel size (mm)
            ('det.dis', 'm'),             # 260  sample to detector distance (m)
            ('det.offset', 'cm'),         # 264  horizontal detector offset (cm)
            ('det.siz', 'cm'),            # 268  physical detector width (cm)
            ('det.bstop', 'mm'),          # 272  beam stop diameter (mm)
            ('resolution.ap1', 'mm'),     # 280  source aperture diameter (mm)
            ('resolution.ap2', 'mm'),     # 284  sample aperture diameter (mm)
            ('resolution.ap12dis', 'm'),  # 288  source aperature to sample aperture distance (m)
            ('resolution.lmda', 'Ang'),   # 292  wavelength (A)
            ('resolution.dlmda', ''),  # 296  wavelength spread (FWHM)
            ('bmstp.xpos', 'cm'),         # 368  beam stop X position (MCU app. cm)
            ('bmstp.ypos', 'cm'),         # 372  beam stop Y position (MCU app. cm)
            ])
        self.precision = { # default precision is 0.01
            }
    def default_value(self, key):
        # type: (str) -> Any
        """
        Return the default value for the field based on field type.
        """
        if self.types[key] == 'vaxR4':
            return 0.
        elif self.types[key] in ('L', 'i'):
            return 0
        elif self.types[key].endswith('s'):
            return ''
        elif self.types[key] == DATETIME_FORMAT:
            return time.gmtime()
        elif self.types[key] == DATE_FORMAT:
            return None
        else:
            raise RuntimeError("Missing default value for "+self.types[key])

INFO = SansFormatInfo(
    ('fname.filename', '21s'),      #   2  filename

    ('run.npre', 'i'),              #  23  number of run prefactors
    ('run.ctime', 'i'),             #  27  count time per prefactor (s)
    ('run.rtime', 'i'),             #  31  total count time (s)
    ('run.numruns', 'i'),           #  35
    ('run.moncnt', 'vaxR4'),        #  39  monitor count
    ('run.savmon', 'vaxR4'),        #  43
    ('run.detcnt', 'vaxR4'),        #  47  detector count (from the anode plane)
    ('run.atten', 'vaxR4'),         #  51  attenuator number (not transmission)
    ('run.datetime', '%d-%b-%Y %H:%M:%S'), #  55  data and time of collection DD-MMM-YYYY HH:MM:SS
    ('run.type', '3s'),             #  75  'RAW'
    ('run.defdir', '11s'),          #  78  NGxSANSnn
    ('run.mode', '1s'),             #  89  C, M
    ('run.reserve', '%d%b%y '),     #  90  DDMMMYY (another date prob. not needed)

    ('sample.labl', '60s'),         #  98  sample label
    ('sample.trns', 'vaxR4'),       # 158  sample transmission
    ('sample.thk', 'vaxR4'),        # 162  sample thickness (cm)
    ('sample.position', 'vaxR4'),   # 166  sample changer position
    ('sample.rotang', 'vaxR4'),     # 170  sample rotation angle (degrees)
    ('sample.table', 'i'),          # 174  chamber or huber position
    ('sample.holder', 'i'),         # 178  sample holder identifier
    ('sample.blank', 'i'),          # 182
    ('sample.temp', 'vaxR4'),       # 186  temperature (sample.tunits)
    # IGOR code says field strength for the electromagnets is at 348
    ('sample.field', 'vaxR4'),      # 190  applied field strength (sample.funits)
    ('sample.tctrlr', 'i'),         # 194  sample control identifier
    ('sample.magnet', 'i'),         # 198  magnet identifier
    ('sample.tunits', '6s'),        # 202  temperature units
    ('sample.funits', '6s'),        # 208  applied field units

    ('det.typ', '6s'),              # 214  ORNL or ILL
    ('det.calx1', 'vaxR4'),         # 220  detector x pixel size (mm)
    ('det.calx2', 'vaxR4'),         # 224  non-linear spatial (10000)
    ('det.calx3', 'vaxR4'),         # 228  corrections (0)
    ('det.caly1', 'vaxR4'),         # 232  detector y pixel size (mm)
    ('det.caly2', 'vaxR4'),         # 236  (10000)
    ('det.caly3', 'vaxR4'),         # 240  (0)
    ('det.num', 'i'),               # 244  area detector identifier
    ('det.spacer', 'i'),            # 248
    ('det.beamx', 'vaxR4'),         # 252  beam center x position (detector coord)
    ('det.beamy', 'vaxR4'),         # 256  beam center y position (detector coord)
    ('det.dis', 'vaxR4'),           # 260  sample to detector distance (m)
    ('det.offset', 'vaxR4'),        # 264  horizontal detector offset (cm)
    ('det.siz', 'vaxR4'),           # 268  physical detector width (cm)
    ('det.bstop', 'vaxR4'),         # 272  beam stop diameter (mm)
    ('det.blank', 'vaxR4'),         # 276

    ('resolution.ap1', 'vaxR4'),    # 280  source aperture diameter (mm)
    ('resolution.ap2', 'vaxR4'),    # 284  sample aperture diameter (mm)
    ('resolution.ap12dis', 'vaxR4'), # 288  source aperature to sample aperture distance (m)
    ('resolution.lmda', 'vaxR4'),   # 292  wavelength (A)
    ('resolution.dlmda', 'vaxR4'),  # 296  wavelength spread (FWHM)
    ('resolution.save', 'vaxR4'),   # 300  lens flag

    ('tslice.slicing', 'L'),        # 304
    ('tslice.multfact', 'i'),       # 308  multiplicative factor for slicing bins
    ('tslice.ltslice', 'i'),        # 312

    ('temp.printemp', 'L'),         # 316  print temp after prefactor
    ('temp.hold', 'vaxR4'),         # 320
    ('temp.err', 'vaxR4'),          # 324
    ('temp.blank', 'vaxR4'),        # 328
    ('temp.extra', 'i'),            # 332  (0x0001 print blue box temp; 0x0100 control from int. bath or ext. probe)
    ('temp.reserve', 'i'),          # 336

    ('magnet.printmag', 'L'),       # 340
    ('magnet.sensor', 'L'),         # 344
    ('magnet.current', 'vaxR4'),    # 348
    ('magnet.conv', 'vaxR4'),       # 352
    ('magnet.fieldlast', 'vaxR4'),  # 356
    ('magnet.blank', 'vaxR4'),      # 360
    ('magnet.spacer', 'vaxR4'),     # 364

    ('bmstp.xpos', 'vaxR4'),        # 368  beam stop X position (MCU app. cm)
    ('bmstp.ypos', 'vaxR4'),        # 372  beam stop Y position (MCU app. cm)

    ('params.blank1', 'i'),         # 376
    ('params.blank2', 'i'),         # 380
    ('params.blank3', 'i'),         # 384
    ('params.trsncnt', 'vaxR4'),    # 388  transmission detector count
    ('params.extra1', 'vaxR4'),     # 392  whole detector transmission
    ('params.extra2', 'vaxR4'),     # 396
    ('params.extra3', 'vaxR4'),     # 400
    ('params.reserve', '42s'),      # 404  first four characters are associated file suffix

    ('voltage.printvolt', 'L'),     # 446
    ('voltage.volts', 'vaxR4'),     # 450 field strength for the superconducting magnet
    ('voltage.blank', 'vaxR4'),     # 454
    ('voltage.spacer', 'i'),        # 458

    ('polarization.printpol', 'L'), # 462
    ('polarization.flipper', 'L'),  # 466
    ('polarization.horiz', 'vaxR4'), # 470
    ('polarization.vert', 'vaxR4'), # 474

    ('analysis.rows1', 'i'),        # 478  box x1  (user defined tranmsission est. box)
    ('analysis.rows2', 'i'),        # 482  box x2
    ('analysis.cols1', 'i'),        # 486  box y1
    ('analysis.cols2', 'i'),        # 490  box y2
    ('analysis.factor', 'vaxR4'),   # 494  box counts
    ('analysis.qmin', 'vaxR4'),     # 498
    ('analysis.qmax', 'vaxR4'),     # 502
    ('analysis.imin', 'vaxR4'),     # 506
    ('analysis.imax', 'vaxR4'),     # 510
)


def load(filename):
    # type: (str) -> Tuple[np.ndarray, Dict[str, Any]]
    """
    Load NCNR VAX SANS format filename.

    Returns data, metadata

    The filename type returned in metadata['run.type'] may be 'RAW' for a
    measurement filename, 'DIV' for a detector sensitivity filename or 'MASK' for
    a detector mask filename.

    The filename name is returned in metadata['fname.filename']

    Raises IOError if the filename length does not correspond to one of the
    possible filename types.  Other errors may be raised if the conversion fails.
    """
    size = os.path.getsize(filename)
    if size == 33316:
        data, metadata = readNCNRData(filename)
    elif size == 66116:
        data = readNCNRSensitivity(filename)
        metadata = {
            'fname.filename': filename,
            'run.type': 'DIV',
            }
    elif size == 16896:
        data = readNCNRMask(filename)
        metadata = {
            'fname.filename': filename,
            'run.type': 'MASK',
            }
    else:
        raise IOError("unknown SANS format filename '%s'"%filename)

    return data, metadata


def save(filename, data, metadata):
    # type: (str, np.ndarray, Dict[str, Any]) -> None
    """
    Save NCNR VAX SANS format filename.

    Input is data, metadata

    metadata['fname.filename'] is not used or modified.

    Raises IOError if the data size does not correspond to the file type.
    """
    writeNCNRData(filename, data, metadata)


def readNCNRSensitivity(inputfile):
    # type: (str) -> np.ndarray
    """
    Read VAX format SANS sensitivity file.
    """
    f = open(inputfile, 'rb')
    data = f.read()
    f.close()

    #skip the fake header and just read the data
    #data is 32bit VAX floats
    dataformatstring = '<65600s'
    #print len(data[516:])
    (rawdatastring,) = struct.unpack_from(dataformatstring, data, offset=516)

    detdata = np.empty(16384)

    a = 0
    offset = 0
    for _i in range(16):
        for _j in range(511):
            #if jj == 0: print rawdatastring[offset:offset+4]
            detdata[a] = R4_VAX2IEEE(rawdatastring[offset:offset+4])
            a += 1
            offset += 4

        offset += 2

        for _k in range(510):
            detdata[a] = R4_VAX2IEEE(rawdatastring[offset:offset+4])
            a += 1
            offset += 4

        offset += 2

    for _i in range(48):
        detdata[a] = R4_VAX2IEEE(rawdatastring[offset:offset+4])
        a += 1
        offset += 4

    detdata.resize((128, 128))

    return detdata


def readNCNRMask(inputfile):
    # type: (str) -> np.ndarray
    """
    Read NCNR format mask file
    """
    f = open(inputfile, 'rb')
    data = f.read()
    f.close()

    # four bytes before and 508 bytes after should be 0
    mask = np.array(struct.unpack_from('<16384B', data, offset=4)) # 4:16384+4
    mask.resize((128, 128))

    return mask

def readNCNRData(inputfile):
    # type: (str) -> Tuple[np.ndarray, Dict[str, Any]]
    """
    Read VAX format SANS data from NCNR.
    """
    #print "reading", inputfile
    f = open(inputfile, 'rb')
    data = f.read()
    f.close()

    if len(data) != 33316:
        raise IOError("Data file corrupted.  Incorrect length")


    # Interpret structure
    #print formatstring
    #print struct.calcsize(HEADER_STRUCT)
    metadata = dict(zip(INFO.fields,
                        struct.unpack_from(INFO.header_struct, data, offset=2)))

    #Process reals into metadata
    for k in INFO.reals:
        metadata[k] = R4_VAX2IEEE(metadata[k])

    #Remove spaces around string fields
    for k in INFO.strings:
        metadata[k] = bytes_to_str(metadata[k]).strip()
    #Convert dates
    metadata['run.datetime'] = time.strptime(bytes_to_str(metadata['run.datetime']),
                                             INFO.types['run.datetime'])

    #print "data len", len(data[514:])
    rawdata = np.array(struct.unpack_from(INFO.data_struct, data, offset=514))

    detdata = decompress(rawdata)

    #print "first 2", struct.unpack_from('>h', data, offset=0)
    return detdata, metadata

def writeNCNRData(filename, data, metadata):
    # type: (str, np.ndarray, Dict[str, Any]) -> None
    """
    Read VAX format SANS data from NCNR.
    """
    # Work with a copy since we are going to update values
    rawdata = metadata.copy()

    # Fill in missing data
    for k in INFO.fields:
        if k not in rawdata:
            rawdata[k] = INFO.default_value(k)

    # convert reals to vax format
    rawdata.update((k, R4_IEEE2VAX(rawdata[k], k)) for k in INFO.reals)

    # round floats to integers rather than truncating
    rawdata.update((k, round(rawdata[k])) for k in INFO.ints)

    # extend strings with spaces
    rawdata.update((k, _raw_str(rawdata[k], INFO.types[k]))
                   for k in INFO.strings)

    # Convert date/time field
    rawdata['run.datetime'] = time.strftime(INFO.types['run.datetime'],
                                            metadata['run.datetime']).upper()

    # If alternate date/time field is not set, set it to date/time value,
    # otherwise treat it as a string and make sure it has 8 characters
    if not rawdata['run.reserve']:
        rawdata['run.reserve'] = time.strftime(INFO.types['run.reserve'],
                                               metadata['run.datetime']).upper()
    else:
        rawdata['run.reserve'] = "%8s"%rawdata['run.reserve']
    rawdata['run.datetime'] = str_to_bytes(rawdata['run.datetime'])
    rawdata['run.reserve'] = str_to_bytes(rawdata['run.reserve'])

    #for k, v in rawdata.items(): print(k, type(v))

    # Pack data into byte arrays
    header = struct.pack(INFO.header_struct, *[rawdata[k] for k in INFO.fields])
    body = struct.pack(INFO.data_struct, *compress(data))  # type: ignore

    #print "reading", inputfile
    with open(filename, 'wb') as f:
        write(f, '\0\0')
        write(f, header)
        write(f, body)

def _raw_str(s, format):
    # type: (bytes, str) -> bytes
    """
    Pad strings with blanks to total length.
    """
    n = int(format[:-1])  # convert format code such as '12s' to length 12
    b = str_to_bytes(s)
    return b[:n] if len(b) >= n else (b+b' '*(n-len(b)))


def decompress(data):
    # type: (np.ndarray) -> np.ndarray
    """
    Convert semi-logarithmic 2-byte integer to 4-byte integer value.

    The storage format for numbers up to 32767 is the number itself
    The storage format for numbers above 32767 is
         - (mantissa + 10000*10**power)
    where the mantissa is 4 digits and power is 1, 2 or 3.
    """

    # Drop values at 0, 1022, 2*1022, ...
    idx = np.arange(len(data), dtype='i')%1022 != 0
    #print "not idx", data[~idx]
    data = data[idx]
    assert len(data) == 16384

    # Logarithmic decompression
    base = 10000
    idx = data <= -base
    power = np.floor(-data[idx]/base)
    data[idx] = np.asarray((-data[idx]%base)*10**power, data.dtype)

    # Recast as 128x128 array
    data = data.reshape((128, 128), order='F')
    return data

def compress(data):
    # type: (np.ndarray) -> np.ndarray
    """
    Convert 4-byte integer value to semi-logarithmic 2-byte integer.

    The storage format for numbers up to 32767 is the number itself
    The storage format for numbers above 32767 is
         - (mantissa + 10000*10**power)
    where the mantissa is 4 digits and power is 1, 2 or 3.

    add an extra integer at 0, 1024
    """
    data = np.asarray(data.flatten(), 'int32')
    assert len(data) == 16384

    # Logarithmic compression
    base = 10000
    erridx = data > 2767000
    idx = data > 32767
    power = np.ceil(np.log10(data[idx]))-4
    mantissa = data[idx] // (10**power)
    data[idx] = np.asarray(-(mantissa + power*base), data.dtype)
    data[erridx] = -777

    # Add values at 0, 1022, 2*1022, ...
    fulldata = np.zeros(16384 + 17, 'i')
    idx = np.arange(len(fulldata), dtype='i')%1022 != 0
    fulldata[idx] = data
    return fulldata

def R4_VAX2IEEE(vax):
    # type: (bytes) -> float
    """
    Convert 4 character VAX REAL*4 string into floating point value
    """
    intValue = bytes_to_int(vax)
    if intValue == 0:
        return 0.
    sign = (intValue >> 15) & 0x1
    exp = (intValue >> 7) & 0xff
    mant = ((intValue & 0x7f) << 16) | ((intValue >> 16) & 0xffff)

    value = (0.5 + mant / float(0x1000000)) * 2**(exp - 128)
    if sign == 1:
        value = -value
    return value

def R4_IEEE2VAX(fpValue, varName):
    # type: (float, str) -> bytes
    """
    Convert floating point value to VAR REAL*4 string
    """
    try:
        m, e = math.frexp(float(fpValue))
        if m == 0.0:
            intValue = 0
        else:
            sign = m < 0
            exp = e + 128
            mant = int((0x1000000 * (abs(m) - 0.5)) + 0.5)
            while mant >= 0x800000:
                exp += 1
                mant = (mant>> 1) - 0x400000
            if exp < 1:
                intValue = 0
            elif exp > 0xff:
                raise Exception("Exponent too large to store as VAX F float for %s."
                                % varName)
            else:
                intValue = ((sign << 15)
                            | (exp << 7)
                            | ((mant & 0x007f0000) >> 16)
                            | ((mant & 0x0000ffff) << 16))
        return int_to_bytes(intValue)
    except:
        print("VAX float F conversion error for %s value: %s"
              % (varName, fpValue))
        return b"\0\0\0\0"

# CRUFT: python 2.x support
if sys.version_info[0] == 2:
    def bytes_to_int(s):
        # type: (str) -> int
        return  ord(s[0]) | (ord(s[1]) << 8) | (ord(s[2]) << 16) | (ord(s[3]) << 24)
    def int_to_bytes(v):
        # type: (int) -> str
        return "".join(chr(k) for k in (v&0xff, (v>>8)&0xff, (v>>16)&0xff, (v>>24)&0xff))
else:
    def bytes_to_int(s):
        # type: (bytes) -> int
        return s[0] | (s[1]<<8) | (s[2]<<16) | (s[3]<<24)
    def int_to_bytes(v):
        # type: (int) -> bytes
        return bytes((v&0xff, (v>>8)&0xff, (v>>16)&0xff, (v>>24)&0xff))


# ==== demo ====
def plot(filename):
    # type: (str) -> None
    """
    Load and plot the SANS data filename.
    """
    from matplotlib import pyplot as plt
    try:
        data, metadata = load(filename)
    except IOError as exc:
        print(exc)
        return

    if metadata['run.type'] == 'RAW':
        data = np.log10(data+1)

    plt.figure()
    plt.imshow(data)
    plt.title(filename)

def _tocvs(f):
    # type: (Any) -> str
    # String
    try:
        return '"%s"'%(f.replace('\000', ' ').replace('\n', '\\n').replace('"', "'").strip())
    except Exception:
        pass
    # Date
    try:
        return time.strftime("%Y-%m-%d %H:%M:%S", f)
    except Exception:
        pass
    # Number
    return str(f)

def csv(fields):
    # type: (Iterator[Any]) -> str
    return ",".join(_tocvs(f) for f in fields)

def expand_input_args(*args):
    # type: (*str) -> List[str]
    """
    Convert args to a list of paths.

    If args[i] is a directory, return all .SA2 and .SA3 files in that directory.
    """
    import glob
    res = []
    for path in args:
        if os.path.isdir(path):
            res.extend(glob.glob(os.path.join(path, '*.SA[23]*')))
        else:
            res.append(path)
    return res


def show_meta(filename):
    # type: (str) -> None
    try:
        _, metadata = readNCNRData(filename)
        # analysis.factor doesn't match counts
        #x1, x2 = metadata['analysis.rows1'], metadata['analysis.rows2']
        #y1, y2 = metadata['analysis.rows1'], metadata['analysis.rows2']
        #metadata['analysis.ROIcounts'] = np.sum(detdata[x1:x2+1, y1:y2+1])
        del metadata['params.reserve']
        for k in sorted(metadata.keys()):
            print("%s: %s"%(k, metadata[k]))
    except IOError:
        print("* could not read %s"%filename)

def show_csv(filename, keys=None, headers=False):
    # type: (str, Optional[Sequence[str]], bool) -> None
    try:
        _, metadata = readNCNRData(filename)
        if keys is None:
            keys = list(sorted(metadata.keys()))
        if headers:
            print(",".join('"%s"'%k for k in keys))
        if 'directory' in keys:
            metadata['directory'] = os.path.dirname(filename)
        print(csv(metadata[k] for k in keys))
    except IOError:
        pass

def show_stats(filename, headers=False):
    # type: (str, bool) -> None
    if headers:
        print("filename, total, duration, rate, max")
    try:
        detdata, metadata = readNCNRData(filename)
        counts = np.sum(detdata)  # type: float
        count_time = metadata['run.rtime']
        rate = counts/count_time
        nmax = np.max(detdata)
        print("%s, %9d, %6d, %7.1f, %7d"
              % (filename, counts, count_time, rate, nmax))
    except Exception:
        print("%s,,,," % filename)

def demo():
    # type: () -> None
    summary_keys = [
        'directory',
        'fname.filename', 'run.datetime',
        'sample.labl', 'sample.position',
        'run.rtime', 'run.atten',
        'det.dis', 'resolution.ap1', 'resolution.ap2',
        'det.bstop', 'bmstp.xpos', 'bmstp.ypos', 'det.beamx', 'det.beamy',
        'resolution.lmda', 'resolution.dlmda',
        'sample.thk', 'sample.trns', 'sample.temp', 'sample.tunits',
        'run.moncnt', 'run.detcnt',
    ]

    if len(sys.argv) == 1 or "-?" in sys.argv:
        print("usage: [-s|-m|-c|-S] file ...")
        print(" -s  stats only")
        print(" -m  show metadata")
        print(" -c  convert to CSV")
        print(" -S  convert to CSV, but only include important fields")
        print(" -?  display this help and exit")
        return

    stats_only = len(sys.argv) > 1 and sys.argv[1] == '-s'
    show_metadata = len(sys.argv) > 1 and sys.argv[1] == '-m'
    show_csvdata = len(sys.argv) > 1 and sys.argv[1] in ('-c', '-S')
    # -S is cvsdata with a subset of the keys
    keys = summary_keys if sys.argv[1] == '-S' else None

    first_file_index = 2 if sys.argv[1].startswith('-') else 1
    files = expand_input_args(*sys.argv[first_file_index:])
    if stats_only:
        first = True
        for f in sys.argv[2:]:
            show_stats(f, headers=first)
            first = False
    elif show_metadata:
        for f in sys.argv[2:]:
            show_meta(f)
    elif show_csvdata:
        first = True
        for f in files:
            show_csv(f, keys=keys, headers=first)
            first = False
    else:
        for f in sys.argv[1:]:
            plot(f)
        from matplotlib import pyplot
        pyplot.show()

def copyfile(infile, outfile):
    # type: (str, str) -> None
    data, metadata = readNCNRData(infile)
    writeNCNRData(outfile, data, metadata)


def main():
    if len(sys.argv) == 4 and sys.argv[1] == "copy":
        copyfile(sys.argv[2], sys.argv[3])
    else:
        demo()

if __name__ == '__main__':
    main()



from reductus.dataflow import core as df
from reductus.dataflow.automod import make_modules

from . import steps
from . import templates
from .dcsdata import RawData, EQData, DCS1dData, EfTwoThetaData

INSTRUMENT = "ncnr.dcs"

def define_instrument():
    # Define modules
    modules = make_modules(steps.ALL_ACTIONS, prefix=INSTRUMENT+'.')

    # Define data types
    dcsraw = df.DataType(INSTRUMENT+".raw", RawData)
    eqdata = df.DataType(INSTRUMENT+".eq", EQData)
    ef2thetadata = df.DataType(INSTRUMENT+".ef2th", EfTwoThetaData)
    eq1ddata = df.DataType(INSTRUMENT+".eq1d", DCS1dData)
    #params = df.DataType(INSTRUMENT+".params", Parameters)

    # Define instrument
    dcs = df.Instrument(
        id=INSTRUMENT,
        name='NCNR Disk Chopper Spectrometer',
        menu=[('steps', modules)],
        datatypes=[dcsraw, eqdata, eq1ddata, ef2thetadata], # params],
        template_defs=df.load_templates(templates),
        )

    # Register instrument
    df.register_instrument(dcs)
    return dcs



from collections import OrderedDict
import datetime

import numpy as np

from .lib.exporters import exports_json

def todict(obj, convert_bytes=False):
    if isinstance(obj, np.integer):
        obj = int(obj)
    elif isinstance(obj, np.floating):
        obj = float(obj)
    elif isinstance(obj, np.ndarray):
        obj = obj.tolist()
    elif isinstance(obj, datetime.datetime):
        obj = [obj.year, obj.month, obj.day, obj.hour, obj.minute, obj.second]
    elif isinstance(obj, (list, tuple)):
        obj = [todict(a, convert_bytes=convert_bytes) for a in obj]
    elif isinstance(obj, (dict, OrderedDict)):
        obj = OrderedDict([(k, todict(v, convert_bytes=convert_bytes))
                           for k, v in obj.items()])
    elif isinstance(obj, bytes) and convert_bytes:
        obj = obj.decode()
    return obj

class Parameters(OrderedDict):
    """
    Untyped parameter object for passing information between modules.

    The source can create the data object using::

        data = Parameters(key=value, ...)

    The target can reference the data object using::

        value = data.key
        value = data['key']

    The order of the key-value pairs will be preserved, and may be displayed
    in that order when the output node is in focus. A key or attribute error
    will be raised if you do not, depending which form you use to access the
    data object.

    If used directly, your modules will need to check that the correct
    data has been passed from the source to the target.

    Alternatively, using a trivial subclass you can provide a typed link
    that will only accept connections from the correct outputs to the
    correct inputs in your graph template.  In your instrument definition
    do the following::

        from reductus.dataflow.data import Parameters
        import dataflow.core as df

        INSTRUMENT = 'ncnr.refl'
        class DeadtimeData(Parameters):
            pass
        deadtime = df.DataType(INSTRUMENT+'.deadtime', DeadtimeData)
        refl1d = df.Instrument(
            id=INSTRUMENT,
            ...
            datatype=[..., deadtime],
        )
        df.register_instrument(refl1d)

    Note: as of python 3.6, the order of the arguments given in the
    constructor will be preserved in the ordered dictionary.
    """
    def get_metadata(self):
        return todict(self)

    def get_plottable(self):
        return {"entry": "entry", "type": "params", "params": todict(self)}

    # For compatibility with existing Parameters objects in sansred, etc.
    @property
    def params(self):
        return self

    # Allow dot-notation to access fields
    def __getattr__(self, name):
        if name in self:
            return self[name]
        raise AttributeError("Name '{name}' is not {self.__class__.__name__}".format(name=name, self=self))

    @exports_json("json")
    def to_json_text(self):
        return {
            "name": getattr(self, "name", self.__class__.__name__),
            #"entry": getattr(self, "entry", "default_entry"),
            "file_suffix": ".json",
            "value": todict(self, convert_bytes=True),
        }

class Plottable:
    """
    Simple class for plotting output that isn't passed from node to node.

    *layout* is an arbitrary plottable.  layout['title'] should be of the
    form "name:entry" in case the user wants to save the output as a json
    structure rather than exporting svg or png.
    """
    def __init__(self, layout):
        self.layout = layout

    def get_metadata(self):
        # Not sure what this is for...
        return todict(self.layout)

    def get_plottable(self):
        return todict(self.layout)

    @exports_json("json")
    def to_json_text(self):
        title = self.layout.get("title", "name:entry")
        name, entry = title.split(':', 1) if ':' in title else (title,"entry")
        return {
            "name": name,
            "entry": entry,
            "file_suffix": ".plot.json",
            "value": todict(self.layout),
        }



import os
import json
from posixpath import basename, join
from copy import copy, deepcopy
from io import BytesIO
import numpy as np
from time import time,strftime

# Action names
__all__ = [] # type: List[str]

# Action methods
ALL_ACTIONS = [] # type: List[Callable[Any, Any]]

def cache(action):
    """
    Decorator which adds the *cached* attribute to the function.

    Use *@cache* to force caching to always occur (for example, when
    the function references remote resources, vastly reduces memory, or is
    expensive to compute.  Use *@nocache* when debugging a function
    so that it will be recomputed each time regardless of whether or not it
    is seen again.
    """
    action.cached = True
    return action

def nocache(action):
    """
    Decorator which adds the *cached* attribute to the function.

    Use *@cache* to force caching to always occur (for example, when
    the function references remote resources, vastly reduces memory, or is
    expensive to compute.  Use *@nocache* when debugging a function
    so that it will be recomputed each time regardless of whether or not it
    is seen again.
    """
    action.cached = False
    return action

def module(action):
    """
    Decorator which records the action in *ALL_ACTIONS*.

    This just collects the action, it does not otherwise modify it.
    """
    ALL_ACTIONS.append(action)
    __all__.append(action.__name__)

    # Sort modules alphabetically
    ALL_ACTIONS.sort(key=lambda action: action.__name__)
    __all__.sort()

    # This is a decorator, so return the original function
    return action

@module
def LoadDCS(filelist=None, check_timestamps=True):
    """
    loads a data file into a SansData obj and returns that.
    Checks to see if data being loaded is 2D; if not, quits

    **Inputs**

    filelist (fileinfo[]): Files to open.
    
    check_timestamps (bool): verify that timestamps on file match request

    **Returns**

    output (raw[]): all the entries loaded.

    2018-04-25 Brian Maranville
    """
    from reductus.dataflow.fetch import url_get
    from .dcsdata import readDCS
    if filelist is None:
        filelist = []
    data = []
    for fileinfo in filelist:
        path, mtime, entries = fileinfo['path'], fileinfo.get('mtime', None), fileinfo.get('entries', None)
        name = basename(path)
        fid = BytesIO(url_get(fileinfo, mtime_check=check_timestamps))
        entry = readDCS(name, fid)
        data.append(entry)

    return data

@module
def RawToEfTwoTheta(rawdata):
    """
    converts raw data to Q,E

    **Inputs**

    rawdata (raw): data in

    **Returns**

    output (ef2th): modified data

    2018-04-29 Daniel Pajerowski
    """
    from io import BytesIO
    from .dcsdata import EfTwoThetaData
    from .dcs_detector_info import detector_info_txt
    detInfo = np.genfromtxt(BytesIO(detector_info_txt.encode()), skip_header=1)#, skip_footer=17)
    detToTwoTheta = detInfo[:,9] # 10th column

    masterSpeed = rawdata.metadata['ch_ms']
    speedRatDenom = rawdata.metadata['ch_srdenom']
    t_SD_min = rawdata.metadata['tsdmin']
    t_channel = np.arange(rawdata.histodata.shape[0])
    ef = Ef_from_timechannel(t_channel, t_SD_min, speedRatDenom, masterSpeed)
    
    new_metadata = deepcopy(rawdata.metadata)
    new_ef2th = EfTwoThetaData(rawdata.name, rawdata.histodata, ef=ef, twotheta=detToTwoTheta, metadata=new_metadata)
    return new_ef2th

@module
def RawToEQ(rawdata):
    """
    converts raw data to Q,E

    **Inputs**

    rawdata (raw): data in

    **Returns**

    output (eq): modified data

    2018-04-28 Daniel Pajerowski
    """
    from io import BytesIO
    from .dcsdata import EQData
    from .dcs_detector_info import detector_info_txt
    detInfo = np.genfromtxt(BytesIO(detector_info_txt.encode()), skip_header=1)#, skip_footer=17)
    detToTwoTheta = detInfo[:,9] # 10th column

    ch_wl = rawdata.metadata['ch_wl']
    Ei = Elam(ch_wl)
    ki = kE(Ei)
    dE = abs(0.5*(-0.10395+0.05616 *Ei+0.00108 *Ei**2)) #take the putative resolution and halve it
    masterSpeed = rawdata.metadata['ch_ms']
    speedRatDenom = rawdata.metadata['ch_srdenom']
    t_SD_min = rawdata.metadata['tsdmin']

    #binning resolution
    Q_max = Qfunc(ki,ki,150)
    Q_min = 0
    E_bins = np.linspace(-Ei, Ei, int(int(2*Ei/dE)*0.5) )
    Q_bins = np.linspace(Q_min,Q_max,int(301*0.5))

    #for every point in {timechannel, detectorchannel} space, map into a bin of {E,Q} space
    #remember, data is organized as data[detectorchannel][timechannel]
    data = rawdata.histodata.T
    i,j = np.indices(data.shape)
    ef = Ef_from_timechannel(j, t_SD_min, speedRatDenom, masterSpeed)
    #print np.shape(data)
    #print np.shape(ki), np.shape(kE(ef)), np.shape(detToTwoTheta[:, None])
    #print detToTwoTheta

    Q_ = Qfunc(ki, kE(ef), detToTwoTheta[:, None])

    E_transfer = Ei-ef
    E_mask = (E_transfer > -Ei)

    EQ_dataarray, xedges, yedges = np.histogram2d(Q_[E_mask], E_transfer[E_mask], bins=(Q_bins, E_bins), range=([Q_min,Q_max], [-Ei, Ei]), weights=data[E_mask])
    # normalize to number of pixels in each histogram bin:
    norm_data = np.ones_like(data)
    EQ_norm, xedges, yedges = np.histogram2d(Q_[E_mask], E_transfer[E_mask], bins=(Q_bins, E_bins), range=([Q_min,Q_max], [-Ei, Ei]), weights=norm_data[E_mask])

    norm_mask = (EQ_norm != 0)
    EQ_normalized = np.copy(EQ_dataarray)
    EQ_normalized[norm_mask] = EQ_dataarray[norm_mask] / EQ_norm[norm_mask]
    
    new_metadata = deepcopy(rawdata.metadata)
    new_metadata['Ei'] = Ei
    new_metadata['Ef'] = ef
    new_metadata['Q_max'] = Q_max
    new_metadata['Q_min'] = Q_min
    new_eq = EQData(rawdata.name, EQ_normalized, new_metadata)
    return new_eq

@module
def sliceEQData(data, slicebox=[None,None,None,None]):
    """
    Sum 2d data along both axes and return 1d datasets

    **Inputs**

    data (eq) : data in
    
    slicebox (range?:xy): region over which to integrate (in data coordinates)

    **Returns**

    xout (eq1d) : xslice

    yout (eq1d) : yslice

    2018-04-22 Brian Maranville
    """

    from .dcsdata import DCS1dData

    if slicebox is None:
        slicebox = [None, None, None, None]
    xmin, xmax, ymin, ymax = slicebox
    
    # then use q-values
    xdata = data.xaxis['values']
    ydata = data.yaxis['values']

    xmin = data.xaxis["min"]
    xmax = data.xaxis["max"]
    x_in = data.xaxis["values"]
    y_in = data.yaxis["values"]
    
    xslice = slice(get_index(x_in, xmin), get_index(x_in, xmax))
    yslice = slice(get_index(y_in, ymin), get_index(y_in, ymax))
    x_out = x_in[xslice]
    y_out = y_in[yslice]
    dx = np.zeros_like(x_out)
    dy = np.zeros_like(y_out)
        
    dataslice = (xslice, yslice)
    x_sum = np.sum(data.data[dataslice], axis=1)
    y_sum = np.sum(data.data[dataslice], axis=0)
    
    x_output = DCS1dData(x_out, x_sum, dx=dx, dv=dx, xlabel=data.xaxis["label"], vlabel="I",
                    xunits="", vunits="neutrons", metadata=data.metadata)
    y_output = DCS1dData(y_out, y_sum, dx=dy, dv=dy, xlabel=data.yaxis["label"], vlabel="I",
                    xunits="", vunits="neutrons", metadata=data.metadata)

    return x_output, y_output

def Elam(lam):
    """
    convert wavelength in angstroms to energy in meV
    """
    return 81.81/lam**2

def Ek(k):
    """
    convert wave-vector in inver angstroms to energy in meV
    """
    return 2.072*k**2

def kE(E):
    return np.sqrt(E/2.072)

def Qfunc(ki, kf, theta):
    """
    evaluate the magnitude of Q from ki, kf, and theta
    theta is the angle between kf and ki, sometimes called 2 theta, units of degrees
    """
    return np.sqrt(  ki**2 + kf**2 - 2*ki*kf*np.cos(theta*np.pi/180)  )

def Ef_from_timechannel(timeChannel, t_SD_min, speedRatDenom, masterSpeed):
    """
    using the parameters
        t_SD_min = minimum sample to detector time
        speedRatDenom = to set FOL chopper speed
        masterSpeed = chopper speed (except for FOL chopper)
    using the variabl
        timeChannel, where I am numbering from 1 <be careful of this convention>
    """
    return 8.41e7 / (t_SD_min + (timeChannel+1)*    (6e4 *(speedRatDenom/masterSpeed))   )**2

def get_index(t, x):
    if (x == "" or x == None):
        return None
    if float(x) > t.max():
        return None
    if float(x) < t.min():
        return None
    tord = np.argsort(t)
    return tord[np.searchsorted(t, float(x), sorter=tord)]


from __future__ import print_function

import re

try:
    import urllib.request as urllib2
except ImportError:
    import urllib2

NIST_DOI_ROOT = "10.18434"
NCNR_DATA_EXT = "T4201B"

NCNR_DATA_DOI = "/".join([NIST_DOI_ROOT, NCNR_DATA_EXT])

def get_DOI_metadata(doi=NCNR_DATA_DOI):
    url = "https://ezid.cdlib.org/id/doi:%s" % (doi,)
    c = urllib2.urlopen(url)
    anvl = c.read()
    metadata = dict(tuple(unescape(v).strip() for v in l.split(":", 1)) \
      for l in anvl.decode("UTF-8").splitlines())
    return metadata

def unescape(s):
    return re.sub("%([0-9A-Fa-f][0-9A-Fa-f])", lambda m: chr(int(m.group(1), 16)), s)

def get_target(doi=NCNR_DATA_DOI):
    # an easier way to get the target value... not tested extensively
    url = "https://dx.doi.org/%s" % (doi,)
    file_pointer = urllib2.urlopen(url)
    return file_pointer.url

if __name__ == '__main__':
    m = get_DOI_metadata()
    print("metadata: %s\n" % m)
    n = get_target()
    print("target: %s" % n)



from reductus.reflred.steps import ncnr_load, subtract_background, divide_intensity, rescale, mask_points, join, divergence, normalize, abinitio_footprint, fit_footprint, correct_footprint

from reductus.dataflow.automod import module


@module
def load(filelist=None,
        intent='auto',
        Qz_basis='detector',
        sample_width=None,
        slit1_distance=None,
        slit1_aperture=None,
        slit2_distance=None,
        slit2_aperture=None
        ):
    r"""
    Load a list of X-ray Reflectivity files.

    *Qz_basis* uses one of the following values:

        **detector**
            ignores the sample angle and calculates Qz
            as $(4\pi/\lambda \sin(\theta_\text{detector}/2))$,
        **sample**
            ignores the detector angle and calculates Qz
            as $(4\pi/\lambda \sin(\theta_\text{sample}))$
        **actual**
            calculates Qx and Qz as (x,z)-components of
            $(\vec k_{\text{out}} - \vec k_\text{in})$ in sample coordinates,

    **Inputs**

    filelist (fileinfo[]): List of files to open.

    intent (opt:auto|specular|background+\|background-\|intensity|rock sample|rock detector|rock qx|scan)
    : Measurement intent (specular, background+, background-, slit, rock),
    auto or infer.  If intent is 'scan', then use the first scanned variable.

    Qz_basis (opt:detector|sample|actual)
    : How to calculate Qz from instrument angles.

    sample_width {Sample width (mm)} (float?)
    : Width of the sample along the beam direction in mm, used for
    calculating the effective resolution when the sample is smaller
    than the beam.  Leave blank to use value from data file.

    slit1_distance {override slit1 distance} (float?)
    : if specified, will override the value found in the file for
    the distance from the sample to slit 1

    slit1_aperture {override slit1 aperture} (float?)
    : if specified, will override the value found in the file for
    the opening of slit 1 in mm

    slit2_distance {override slit2 distance} (float?)
    : if specified, will override the value found in the file for
    the distance from the sample to slit 2

    slit2_aperture {override slit2 aperture} (float?)
    : if specified, will override the value found in the file for
    the opening of slit 2 in mm

    **Returns**

    output (refldata[]): All entries of all files in the list.

    | 2022-01-31 Brian Maranville
    | 2022-04-27 Brian Maranville fix count time for Rigaku in sweep mode
    """
    from reductus.reflred.load import url_load_list
    from reductus.reflred import xrawref
    
    # TODO: sample_width is ignored if datafile defines angular_divergence
    auto_divergence = True
    enforce_specular = True

    datasets = []
    for data in url_load_list(filelist, loader=xrawref.load_entries):
        data.Qz_basis = Qz_basis
        if intent not in [None, 'auto']:
            data.intent = intent
        if data.intent == 'specular' and enforce_specular:
            if data.Qz_basis == 'detector':
                data.sample.angle_x = data.sample.angle_x_target = data.detector.angle_x / 2.0
            elif data.Qz_basis == 'sample':
                data.detector.angle_x = data.detector.angle_x_target = data.sample.angle_x * 2.0
        if slit1_distance is not None:
            data.slit1.distance = slit1_distance
        if slit1_aperture is not None:
            data.slit1.x = data.slit1.x_target = slit1_aperture
        if slit2_distance is not None:
            data.slit2.distance = slit2_distance
        if slit2_aperture is not None:
            data.slit2.x = data.slit2.x_target = slit2_aperture
        if auto_divergence:
            data = divergence(data, sample_width)
        
        data = normalize(data, base='time')
        #print "data loaded and normalized"
        datasets.append(data)

    return datasets
